/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      ComboBox
Version:   1.0
ID:        ID_COMBOBOX
Status:    Beta
Category:  GUI
Date:      December 2003
Author:    Rocklyte Systems
Copyright: Rocklyte Systems, 2003.  All rights reserved.
Keywords:  combobox, combo, box, text, entry, list
Short:     The ComboBox class manages the display and interaction of user combo boxes.

###DESCRIPTION###
<p>The ComboBox class is used to create combo boxes, also known as 'drop-down
menus' in application interfaces.  A combo box typically looks like a text
entry area, but features a button positioned to the right-hand side of the gadget.
Clicking on the button will pop-up a menu that the user can use to select a
pre-defined menu item.  Clicking on one of those items will paste the item text
into the combo box.</p>

<p>Definitions for the combobox are loaded by default from the environment
file "templates:comboboxdef.xml".  You can change the template file prior to
initialisation by setting the <a href="#Template">Template</a> field.  Note
that the values set in the template will override the original field
settings for the combobox.</p>

<p>The most important feature of the combobox is the drop-down menu.  The
combobox uses the <a href="menu.html">Menu</a> class to support its menu construction.  To add items
to the drop-down menu, you need to pass instructions to it using XML.  You can
find out more about this XML specification in the documentation for the <a href="menu.html">Menu</a>
class.  The following example illustrates the simplest way to create a combobox
with a pre-defined set of items:</p>

<pre>
  &lt;combobox label="Options:" x="[owner.leftmargin]" y="[owner.topmargin]"
    width="200"&gt;
    &lt;xml&gt;
      &lt;item text="Tamarillo"/"&gt;
      &lt;item text="Kumquat"/"&gt;
      &lt;item text="Mango"/"&gt;
    &lt;/xml&gt;
  &lt;/combobox&gt;
</pre>

<p>When the user selects a combobox item, you may need to respond with an
action.  You can do this by initialising child objects to the combobox.
These will be executed when the combobox is activated.  You can also set the
<a href="#ActionScript">ActionScript</a> field to run a script on
activation.  When programming, you can also subscribe to the combobox's
Activate action and write a customised response routine.</p>

###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/files/all.h>
#include <pandora/misc/all.h>
#include <pandora/modules/strings.h>
#include <pandora/modules/filesystem.h>
#include <pandora/input/joyport.h>
#include <pandora/input/keyboard.h>
#include <pandora/data/xml.h>
#include <pandora/functions/set.c>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_COMBOBOX  1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_COMBOBOX, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2003. All rights reserved.",
   "December 2003",
   "ComboBox"
};

static struct KernelBase  *KernelBase;
static struct StringsBase *StringsBase;
static struct FileSystemBase *FileSystemBase;
static OBJECTPTR ComboBoxClass = NULL;
static OBJECTPTR StringsMod = NULL;
static OBJECTPTR FileSystemMod = NULL;

/*****************************************************************************
** Object structure.
*/

struct ComboBox {
   OBJECT_HEADER              /* Class header */
   OBJECTID  RegionID;        /* Drawable region created by the combobox object */
   OBJECTID  DrawableID;      /* The drawable target for the combobox graphic */
   LONG      Flags;           /* Special options */
   struct    Font *Font;      /* Font control object - for the label graphic */
   OBJECTPTR Text;            /* Text control object - for the combobox area */
   OBJECTPTR Menu;            /* Menu control object */
   LONG      FocusFrame;      /* Frame to display when the combobox has the focus */
   LONG      ReleaseFrame;    /* Frame to display when the combobox loses the focus */
   LONG      Thickness;       /* Border thickness */
   LONG      LabelWidth;      /* Width of the combobox label area */
   OBJECTID  ButtonID;        /* Drawable ID of the combobox button */
   LONG      ButtonOffset;    /* Offset to apply to the button position from the border areas */

   PRIVATE_FIELDS
   struct RGB ColourRGB;      /* Colour to use as the background */
   struct RGB HighlightRGB;   /* Colour to use for border highlighting */
   struct RGB ShadowRGB;      /* Colour to use for border shadowing */
   UBYTE     Label[50];       /* Label */
   OBJECTPTR TemplateXML;
   OBJECTPTR XML;
   OBJECTID  TextID;
   OBJECTID  FontID;
   OBJECTID  MenuID;
   MEMORYID  ActionScriptMID;
   STRING    ActionScript;     /* Simple DML statement indicating what to do on combobox-press, e.g. "<set object="[desktop]" field="">"... */
   BYTE      State;
   BYTE      Active;
};

/*****************************************************************************
** Class definition.
*/

static struct FieldDef ComboBoxFlags[] = {
   { "HIDE",          CMF_HIDE },
   { "DISABLED",      CMF_DISABLED },
   { "NOBKGD",        CMF_NOBKGD },
   { "LOCAL",         CMF_LOCAL },
   { "FOCUSACTIVATE", CMF_FOCUSACTIVATE },
   { "LIMITTOLIST",   CMF_LIMITTOLIST },
   { "EDIT",          CMF_EDIT },
   { "AUTOCOMPLETE",  CMF_AUTOCOMPLETE },
   { NULL, NULL }
};

static ERROR GET_ActionScript(struct ComboBox *, STRING *);
static ERROR GET_Bottom(struct ComboBox *, LONG *);
static ERROR GET_ColourRGB(struct ComboBox *, struct RGB *);
static ERROR GET_Disable(struct ComboBox *, LONG *);
static ERROR GET_FocusActivate(struct ComboBox *, LONG *);
static ERROR GET_Height(struct ComboBox *, struct Variable *);
static ERROR GET_HighlightRGB(struct ComboBox *, struct RGB *);
static ERROR GET_Label(struct ComboBox *, STRING *);
static ERROR GET_Right(struct ComboBox *, LONG *);
static ERROR GET_ShadowRGB(struct ComboBox *, struct RGB *);
static ERROR GET_String(struct ComboBox *, STRING *);
static ERROR GET_Width(struct ComboBox *, struct Variable *);
static ERROR GET_X(struct ComboBox *, struct Variable *);
static ERROR GET_XOffset(struct ComboBox *, struct Variable *);
static ERROR GET_Y(struct ComboBox *, struct Variable *);
static ERROR GET_YOffset(struct ComboBox *, struct Variable *);

static ERROR SET_ActionScript(struct ComboBox *, STRING);
static ERROR SET_Border(struct ComboBox *, STRING);
static ERROR SET_Colour(struct ComboBox *, STRING);
static ERROR SET_Disable(struct ComboBox *, LONG *);
static ERROR SET_FocusActivate(struct ComboBox *, LONG *);
static ERROR SET_Height(struct ComboBox *, struct Variable *);
static ERROR SET_Highlight(struct ComboBox *, STRING);
static ERROR SET_Label(struct ComboBox *, STRING);
static ERROR SET_Region(struct ComboBox *, LONG *);
static ERROR SET_Shadow(struct ComboBox *, STRING);
static ERROR SET_String(struct ComboBox *, STRING);
static ERROR SET_TabFocus(struct ComboBox *, OBJECTID *);
static ERROR SET_Template(struct ComboBox *, STRING);
static ERROR SET_Width(struct ComboBox *, struct Variable *);
static ERROR SET_X(struct ComboBox *, struct Variable *);
static ERROR SET_XOffset(struct ComboBox *, struct Variable *);
static ERROR SET_Y(struct ComboBox *, struct Variable *);
static ERROR SET_YOffset(struct ComboBox *, struct Variable *);

static struct FieldArray ComboBoxFields[] = {
   { "Region",         0, FDF_OBJECTID|FDF_RW,  ID_RENDER, NULL, SET_Region },
   { "Drawable",       0, FDF_OBJECTID|FDF_RW,  ID_RENDER, NULL, NULL },
   { "Flags",          0, FDF_LONGFLAGS|FDF_RW, (LONG)&ComboBoxFlags, NULL, NULL },
   { "Font",           0, FDF_CHILD|FDF_R,      0, NULL, NULL },
   { "TextInput",      0, FDF_CHILD|FDF_R,      0, NULL, NULL },
   { "Menu",           0, FDF_CHILD|FDF_R,      0, NULL, NULL },
   { "FocusFrame",     0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "ReleaseFrame",   0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "Thickness",      0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "LabelWidth",     0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   { "Button",         0, FDF_OBJECTID|FDF_RI,  0, NULL, NULL },
   { "ButtonOffset",   0, FDF_LONG|FDF_RW,      0, NULL, NULL },
   /** Virtual fields ***/
   { "ActionScript",   0, FDF_STRING|FDF_RW,   0, GET_ActionScript,  SET_ActionScript },
   { "Border",         0, FDF_STRING|FDF_W,    0, NULL,              SET_Border },
   { "Bottom",         0, FDF_LONG|FDF_R,      0, GET_Bottom,        NULL },
   { "Disable",        0, FDF_LONG|FDF_RW,     0, GET_Disable,       SET_Disable },
   { "Colour",         0, FDF_STRING|FDF_W,    0, NULL,              SET_Colour },
   { "ColourRGB",      0, FDF_POINTER|FDF_R,   0, GET_ColourRGB,     NULL },
   { "FocusActivate",  0, FDF_LONG|FDF_RW,     0, GET_FocusActivate, SET_FocusActivate },
   { "Highlight",      0, FDF_STRING|FDF_W,    0, NULL,              SET_Highlight },
   { "HighlightRGB",   0, FDF_POINTER|FDF_R,   0, GET_HighlightRGB,  NULL },
   { "Label",          0, FDF_STRING|FDF_RW,   0, GET_Label,         SET_Label },
   { "Right",          0, FDF_LONG|FDF_R,      0, GET_Right,         NULL },
   { "Shadow",         0, FDF_STRING|FDF_W,    0, NULL,              SET_Shadow },
   { "ShadowRGB",      0, FDF_POINTER|FDF_R,   0, GET_ShadowRGB,     NULL },
   { "String",         0, FDF_STRING|FDF_RW,   0, GET_String,        SET_String },
   { "TabFocus",       0, FDF_OBJECTID|FDF_I,  ID_TABFOCUS, NULL,    SET_TabFocus },
   { "Template",       0, FDF_STRING|FDF_I,    0, NULL,              SET_Template },
   { "Text",           0, FDF_STRING|FDF_RW,   0, GET_String,        SET_String },
   /*** Variable Fields ***/
   { "Height",         0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Height,  SET_Height },
   { "Width",          0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Width,   SET_Width },
   { "X",              0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XCoord",         0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XOffset",        0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_XOffset, SET_XOffset },
   { "Y",              0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YCoord",         0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YOffset",        0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_YOffset, SET_YOffset },
   END_FIELD
};

/*****************************************************************************
** Actions
*/

static ERROR COMBOBOX_AccessObject(struct ComboBox *, APTR);
static ERROR COMBOBOX_ActionNotify(struct ComboBox *, struct acActionNotify *);
static ERROR COMBOBOX_Activate(struct ComboBox *, APTR);
static ERROR COMBOBOX_ClosingTag(struct ComboBox *, APTR);
static ERROR COMBOBOX_DataChannel(struct ComboBox *, struct acDataChannel *);
static ERROR COMBOBOX_Disable(struct ComboBox *, APTR);
static ERROR COMBOBOX_Enable(struct ComboBox *, APTR);
static ERROR COMBOBOX_Focus(struct ComboBox *, APTR);
static ERROR COMBOBOX_Free(struct ComboBox *, APTR);
static ERROR COMBOBOX_Hide(struct ComboBox *, APTR);
static ERROR COMBOBOX_Init(struct ComboBox *, APTR);
static ERROR COMBOBOX_MoveToBack(struct ComboBox *, APTR);
static ERROR COMBOBOX_MoveToFront(struct ComboBox *, APTR);
static ERROR COMBOBOX_NewObject(struct ComboBox *, APTR);
static ERROR COMBOBOX_Redimension(struct ComboBox *, struct acRedimension *);
static ERROR COMBOBOX_ReleaseObject(struct ComboBox *, APTR);
static ERROR COMBOBOX_Resize(struct ComboBox *, struct acResize *);
static ERROR COMBOBOX_SetUnlistedField(struct ComboBox *, struct acSetUnlistedField *);
static ERROR COMBOBOX_Show(struct ComboBox *, APTR);

static struct ActionArray ComboBoxActions[] = {
   { AC_AccessObject,     COMBOBOX_AccessObject },
   { AC_ActionNotify,     COMBOBOX_ActionNotify },
   { AC_Activate,         COMBOBOX_Activate },
   { AC_ClosingTag,       COMBOBOX_ClosingTag },
   { AC_DataChannel,      COMBOBOX_DataChannel },
   { AC_Disable,          COMBOBOX_Disable },
   { AC_Enable,           COMBOBOX_Enable },
   { AC_Focus,            COMBOBOX_Focus },
   { AC_Free,             COMBOBOX_Free },
   { AC_Hide,             COMBOBOX_Hide },
   { AC_Init,             COMBOBOX_Init },
   { AC_MoveToBack,       COMBOBOX_MoveToBack },
   { AC_MoveToFront,      COMBOBOX_MoveToFront },
   { AC_NewObject,        COMBOBOX_NewObject },
   { AC_Redimension,      COMBOBOX_Redimension },
   { AC_ReleaseObject,    COMBOBOX_ReleaseObject },
   { AC_Resize,           COMBOBOX_Resize },
   { AC_SetUnlistedField, COMBOBOX_SetUnlistedField },
   { AC_Show,             COMBOBOX_Show },
   { NULL, NULL }
};

/*****************************************************************************
** Prototypes
*/

static void DrawComboBox(struct ComboBox *, struct Render *, struct Bitmap *);

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   /*** Open required modules ***/

   if (CreateObject(ID_MODULE, NULL, &StringsMod, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(StringsMod, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &FileSystemMod, NULL,
      FID_Name|TSTRING,   "filesystem",
      FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
      TAGEND) IS ERR_Okay) {
      if (GetField(FileSystemMod, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&ComboBoxClass, NULL,
      FID_BaseClassID|TLONG, ID_COMBOBOX,
      FID_Version|TFLOAT,    VER_COMBOBOX,
      FID_Name|TSTRING,      "ComboBox",
      FID_Category|TLONG,    CCF_GUI,
      FID_Flags|TLONG,       CLF_PROMOTECHILDREN,
      FID_Actions|TPTR,      ComboBoxActions,
      FID_Fields|TPTR,       ComboBoxFields,
      FID_Size|TLONG,        sizeof(struct ComboBox),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (ComboBoxClass) { Action(AC_Free, ComboBoxClass, NULL);   ComboBoxClass = NULL; }
   if (StringsMod)    { Action(AC_Free, StringsMod, NULL);    StringsMod = NULL; }
   if (FileSystemMod) { Action(AC_Free, FileSystemMod, NULL); FileSystemMod = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** ComboBox: AccessObject
*/

static ERROR COMBOBOX_AccessObject(struct ComboBox *Self, APTR Void)
{
   if (Self->FontID) {
      if (AccessObject(Self->FontID, 0, (OBJECTPTR *)&Self->Font) != ERR_Okay) {
         return(ObjectError(Self, ERH_AccessObject, ERR_AccessObject));
      }
   }

   if (Self->TextID) {
      if (AccessObject(Self->TextID, 0, (OBJECTPTR *)&Self->Text) != ERR_Okay) {
         return(ObjectError(Self, ERH_AccessObject, ERR_AccessObject));
      }
   }

   if (Self->MenuID) {
      if (AccessObject(Self->MenuID, 0, (OBJECTPTR *)&Self->Menu) != ERR_Okay) {
         return(ObjectError(Self, ERH_AccessObject, ERR_AccessObject));
      }
   }
   
   return(ERR_Okay);
}

/*****************************************************************************
** ComboBox: ActionNotify
*/

static ERROR COMBOBOX_ActionNotify(struct ComboBox *Self, struct acActionNotify *Args)
{
   if (Args->Error != ERR_Okay) return(ERR_Okay);

   if ((Args->ActionID IS AC_Activate) AND (Args->ObjectID IS Self->ButtonID)) {
      /*** The button on the combo box has been pressed, so display the menu ***/
      Action(AC_Activate, Self->Menu, NULL);
   }
   else if ((Args->ActionID IS AC_Activate) AND (Args->ObjectID IS Self->TextID)) {
      /*** The user has entered something into the text box area ***/
      DPrintF("~ActionNotify()","[ComboBox] The Text object has been used - activating self.");
         Action(AC_Activate, Self, NULL);
      StepBack();
   }
   else if (Args->ActionID IS AC_UserClick) {
      struct acUserClick *click;
      LONG flags;

      click = (struct acUserClick *)Args->Args;

      if (click->XCoord >= Self->LabelWidth) {
         GetField(Self->Text, FID_Flags, FT_LONG, &flags);
         if (!(flags & TXF_EDIT)) Action(AC_Activate, Self->Menu, NULL);
      }
   }
   else if (Args->ActionID IS AC_Redimension) {
      struct acRedimension *redimension;

      redimension = (struct acRedimension *)Args->Args;
      SetFields(Self->Menu, FID_Width|TLONG, (LONG)(redimension->Width - Self->LabelWidth), TAGEND);
   }
   else if (Args->ActionID IS AC_Disable) {
      Self->Flags |= CMF_DISABLED;
      DelayMsg(AC_Draw, Self->RegionID, NULL);
   }
   else if (Args->ActionID IS AC_Enable) {
      Self->Flags &= ~CMF_DISABLED;
      DelayMsg(AC_Draw, Self->RegionID, NULL);
   }
   else if (Args->ActionID IS AC_LostFocus) {
      acHide(Self->Menu);
   }
   else return(ERR_NoSupport);

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Activate
Short: Activates the combobox object's response mechanism.
###END###

*****************************************************************************/

static ERROR COMBOBOX_Activate(struct ComboBox *Self, APTR Void)
{
   STRING script;
   LONG count;
   WORD i;

   DPrintF("~Activate()","[ComboBox:%d]", Self->Head.UniqueID);

   if (Self->Active IS TRUE) {
      DPrintF("@Activate:","Warning - recursion detected");
      StepBack();
      return(ERR_Failed);
   }

   Self->Active = TRUE;

   if (GET_ActionScript(Self, &Self->ActionScript) IS ERR_Okay) {
      if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
         FID_Target|TLONG,     Self->RegionID,
         FID_Location|TSTRING, Self->ActionScript,
         TAGEND) IS ERR_Okay) {
         Action(AC_Activate, script, NULL);
         Action(AC_Free, script, NULL);
      }
   }

   if ((count = TotalChildren(Self->Head.UniqueID))) {
      struct ChildEntry list[count];
      if (ListChildren(Self->Head.UniqueID, list, &count) IS ERR_Okay) {
         if (Self->Flags & CMF_LOCAL) {
            for (i=0; i < count; i++) ActionMsg(AC_Activate, list[i].ObjectID, NULL);
         }
         else for (i=0; i < count; i++) DelayMsg(AC_Activate, list[i].ObjectID, NULL);
      }
   }

   Self->Active = FALSE;

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** ComboBox: ClosingTag
*/

static ERROR COMBOBOX_ClosingTag(struct ComboBox *Self, APTR Void)
{
   if (!(Self->Flags & CMF_HIDE)) Action(AC_Show, Self, NULL);
   return(ERR_Okay);
}

/*****************************************************************************
** ComboBox: DataChannel
*/

static ERROR COMBOBOX_DataChannel(struct ComboBox *Self, struct acDataChannel *Args)
{
   if (!Args) return(ObjectError(Self, ERH_DataChannel, ERR_Args));

   if ((Args->DataType IS DATA_KEYPRESS) AND (Args->Version IS 1)) {
      struct dcKeyEntry *key;
      WORD i;

      if ((key = (struct dcKeyEntry *)Args->Buffer)) {
         for (i=0; i < Args->TotalEntries; i++) {
            if (key[i].Flags & KQ_PRESSED) {
               if ((key[i].Value IS K_ENTER) OR (key[i].Value IS K_NPENTER)) {
                  DPrintF("~DataChannel()","[Text] Enter key detected.");
                     Action(AC_Activate, Self, NULL);
                  StepBack();
               }
            }
         }
      }
   }
   else if ((Args->DataType IS DATA_XML) AND (Args->Version IS 1)) {
      Action(AC_DataChannel, Self->Menu, Args);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Disable
Short: Turns the combobox off.
###END###

*****************************************************************************/

static ERROR COMBOBOX_Disable(struct ComboBox *Self, APTR Void)
{
   /* See the ActionNotify routine to see what happens
   ** when the drawable is disabled.
   */
   DPrintF("Disable()","[ComboBox:%d]", Self->Head.UniqueID);
   ActionMsg(AC_Disable, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Enable
Short: Turns the combobox back on if it has previously been disabled.
###END###

*****************************************************************************/

static ERROR COMBOBOX_Enable(struct ComboBox *Self, APTR Void)
{
   /* See the ActionNotify routine to see what happens
   ** when the drawable is enabled.
   */
   DPrintF("Enable()","[ComboBox:%d]", Self->Head.UniqueID);
   ActionMsg(AC_Enable, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Focus
Short: Sets the focus on the combobox.
###END###

*****************************************************************************/

static ERROR COMBOBOX_Focus(struct ComboBox *Self, APTR Void)
{
   if (Self->Flags & CMF_EDIT) ActionMsg(AC_Focus, Self->RegionID, NULL);
   else ActionMsg(AC_Focus, Self->ButtonID, NULL);

   return(ERR_Okay);
}

/*****************************************************************************
** ComboBox: Free
*/

static ERROR COMBOBOX_Free(struct ComboBox *Self, APTR Void)
{
   if (Self->ActionScriptMID) {
      if (Self->ActionScript) {
         ReleaseMemory(Self->ActionScript);
         Self->ActionScript = NULL;
      }
      FreeMemoryID(Self->ActionScriptMID);
      Self->ActionScriptMID = NULL;
   }

   if (Self->XML) {
      Action(AC_Free, Self->XML, NULL);
      Self->XML = NULL;
   }

   if (Self->ButtonID) {
      ActionMsg(AC_Free, Self->ButtonID, NULL);
      Self->ButtonID = NULL;
   }

   if (Self->FontID) {
      if (Self->Font) {
         Action(AC_Free, Self->Font, NULL);
         ReleaseObject(Self->Font);
         Self->Font = NULL;
      }
      else ActionMsg(AC_Free, Self->FontID, NULL);
      Self->FontID = NULL;
   }

   if (Self->TextID) {
      if (Self->Text) {
         Action(AC_Free, Self->Text, NULL);
         ReleaseObject(Self->Text);
         Self->Text = NULL;
      }
      else ActionMsg(AC_Free, Self->TextID, NULL);
      Self->TextID = NULL;
   }

   if (Self->MenuID) {
      if (Self->Menu) {
         Action(AC_Free, Self->Menu, NULL);
         ReleaseObject(Self->Menu);
         Self->Menu = NULL;
      }
      else ActionMsg(AC_Free, Self->MenuID, NULL);
      Self->MenuID = NULL;
   }

   if (Self->RegionID) {
      ActionMsg(AC_Free, Self->RegionID, NULL);
      Self->RegionID = NULL;
   }

   if (Self->TemplateXML) {
      Action(AC_Free, Self->TemplateXML, NULL);
      Self->TemplateXML = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Hide
Short: Removes the combobox from the display.
###END###

*****************************************************************************/

static ERROR COMBOBOX_Hide(struct ComboBox *Self, APTR Void)
{
   ActionMsg(AC_Hide, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************
** ComboBox: Init()
*/

static ERROR COMBOBOX_Init(struct ComboBox *Self, APTR Void)
{
   struct Render *region;
   OBJECTPTR object;
   OBJECTID owner_id;
   LONG flags, region_height, region_width;

   /*** Find our parent drawable ***/

   if (!Self->DrawableID) {
      owner_id = GetOwner(Self);
      while ((owner_id) AND (GetClassID(owner_id) != ID_RENDER)) {
         owner_id = GetOwnerID(owner_id);
      }
      if (owner_id) Self->DrawableID = owner_id;
      else return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
   }

   /*** Initialise the Font ***/

   if (Action(AC_Init, Self->Font, NULL) != ERR_Okay) {
      return(ObjectError(Self, ERH_Init, ERR_Init));
   }

   /*** Calculate the width of the text label, if there is one ***/

   if ((Self->LabelWidth < 1) AND (Self->Label[0])) {
      struct mtCalcStringWidth calc;
      calc.String = Self->Label;
      calc.Length = -1;
      Action(MT_CalcStringWidth, Self->Font, &calc);
      Self->LabelWidth = calc.Width + 4;
   }

   /*** Initialise the combobox region ***/

   if (AccessObject(Self->RegionID, 5000, (OBJECTPTR *)&region) IS ERR_Okay) {
      SetFields(region, FID_Parent|TLONG, Self->DrawableID,
                        FID_Region|TLONG, TRUE,
                        TAGEND);

      if (!(region->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) {
         if ((!(region->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y))) OR (!(region->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET)))) {
            SetFields(region, FID_Height|TLONG, Self->Font->MaxHeight + ((Self->Thickness+2)*2), TAGEND);
         }
      }

      if (Action(AC_Init, region, NULL) IS ERR_Okay) {
         SubscribeActionTags(region, Self,
            AC_Disable,
            AC_Enable,
            AC_LostFocus,
            AC_Redimension,
            AC_UserClick,
            TAGEND);

         region_width  = region->Width;
         region_height = region->Height;
      }
      else {
         ReleaseObject(region);
         return(ObjectError(Self, ERH_Init, ERR_Init));
      }

      ReleaseObject(region);
   }
   else return(ObjectError(Self, ERH_Init, ERR_AccessObject));

   /*** Use the base template to create the combobox graphics ***/

   if (Self->TemplateXML) {
      OBJECTPTR script;
      struct mtGetXMLString getxml;
      struct XMLTag **xmllist, *tag;

      if (GetFields(Self->TemplateXML, FID_Tags|TPTR, &xmllist, TAGEND) IS ERR_Okay) {
         for (tag=xmllist[0]; tag; tag=tag->Next) {
            if (tag->Child) {
               if (StrCompare("graphics", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                  getxml.Index = tag->Child->Index;
                  getxml.Flags = XMF_INCLUDESIBLINGS;
                  Action(MT_GetXMLString, Self->TemplateXML, &getxml);

                  if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
                     FID_String|TSTRING, getxml.Result,
                     FID_Target|TLONG,   Self->RegionID,
                     TAGEND) IS ERR_Okay) {

                     if (Action(AC_Activate, script, NULL) IS ERR_Okay) {
                        Self->Flags |= CMF_NOBKGD;
                     }

                     Action(AC_Free, script, NULL);
                  }

                  FreeMemory(getxml.Result);
                  break;
               }
            }
         }
      }

      Action(AC_Free, Self->TemplateXML, NULL);
      Self->TemplateXML = NULL;
   }

   if (AccessObject(Self->RegionID, 5000, (OBJECTPTR *)&region) IS ERR_Okay) {
      ActionTags(MT_SetRenderCallback, region, Self, &DrawComboBox);
      ReleaseObject(region);
   }
   else return(ERR_AccessObject);

   /* Initialise the text area that the user will be
   ** able to interact with.
   */

   flags = TXF_ENTERACTIVATE;
   if (Self->Flags & CMF_EDIT) flags |= TXF_EDIT;
   if (Self->Flags & CMF_FOCUSACTIVATE) flags |= TXF_FOCUSACTIVATE;

   SetFields(Self->Text, FID_Drawable|TLONG,    Self->RegionID,
                         FID_Flags|TLONG,       flags,
                         FID_Face|TSTRING,      Self->Font->Face,
                         FID_Point|TLONG,       Self->Font->Point,
                         FID_XCoord|TLONG,      Self->LabelWidth + Self->Thickness + 4,
                         FID_YCoord|TLONG,      Self->Thickness,
                         FID_XOffset|TLONG,     Self->Thickness,
                         FID_YOffset|TLONG,     Self->Thickness,
                         FID_LineLimit|TLONG,   1,
                         FID_RightMargin|TLONG, Self->Thickness + 4,
                         FID_Align|TLONG,       ALIGN_VERTICAL,
                         TAGEND);

   if (Action(AC_Init, Self->Text, NULL) IS ERR_Okay) {
      SubscribeActionTags(Self->Text, Self, AC_Activate, TAGEND);
   }
   else return(ObjectError(Self, ERH_Init, ERR_Init));

   SetFields(Self->Menu,
      FID_Relative|TLONG,         Self->RegionID,
      FID_XCoord|TLONG,           Self->LabelWidth,
      FID_YCoord|TLONG,           region_height - 1,
      FID_VSpacing|TLONG,         4,
      FID_MenuFile|TSTRING,       "templates:dropdowndef.xml",
      FID_Width|TLONG,            region_width - Self->LabelWidth,
      FID_SelectionObject|TLONG,  Self->Head.UniqueID,
      FID_SelectionField|TSTRING, "String",
      FID_Face|TSTRING,           Self->Font->Face,
      FID_Point|TLONG,            Self->Font->Point,
      FID_Flags|TLONG,            MNF_IGNOREFOCUS,
      TAGEND);

   if (Action(AC_Init, Self->Menu, NULL) != ERR_Okay) return(ERR_Init);

   if (!Self->ButtonID) {
      if (CreateObject(ID_BUTTON, NULL, &object, &Self->ButtonID,
         FID_Owner|TLONG,   Self->RegionID,
         FID_XOffset|TLONG, Self->ButtonOffset,
         FID_Y|TLONG,       Self->ButtonOffset,
         FID_YOffset|TLONG, Self->ButtonOffset,
         FID_Width|TLONG,   region_height - (Self->ButtonOffset * 2),
         TAGEND) IS ERR_Okay) {

         SubscribeActionTags(object, Self, AC_Activate, TAGEND);

         ReleaseObject(object);

         LONG arrowwidth, arrowheight, size;

         size = region_height - (Self->ButtonOffset * 2);
         if (size > 32) {
            arrowwidth = 11;
            arrowheight = 6;
         }
         else if (size > 16) {
            arrowwidth = 9;
            arrowheight = 5;
         }
         else {
            arrowwidth = 7;
            arrowheight = 4;
         }

         if (CreateObject(ID_ARROW, NULL, NULL, NULL,
            FID_Owner|TLONG,       Self->ButtonID,
            FID_XOffset|TLONG,     0,
            FID_YCoord|TLONG,      0,
            FID_Width|TLONG,       size,
            FID_Height|TLONG,      size,
            FID_ArrowWidth|TLONG,  arrowwidth,
            FID_ArrowHeight|TLONG, arrowheight,
            FID_Align|TLONG,       ALIGN_CENTER,
            FID_Direction|TSTRING, "down",
            FID_Colour|TSTRING,    "90,90,90",
            TAGEND) IS ERR_Okay) {

         }
         else {
            DPrintF("@Init:","[ComboBox:%d] Failed to create Arrow.", Self->Head.UniqueID);
            return(ERR_CreateObject);
         }

         ActionMsg(AC_Show, Self->ButtonID, NULL);
      }
      else {
         DPrintF("@Init:","[ComboBox:%d] Failed to create Button.", Self->Head.UniqueID);
         return(ERR_CreateObject);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  MoveToBack
Short: Moves the combobox to the back of the display area.
###END###

*****************************************************************************/

static ERROR COMBOBOX_MoveToBack(struct ComboBox *Self, APTR Void)
{
   ActionMsg(AC_MoveToBack, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  MoveToFront
Short: Moves the combobox to the front of the display area.
###END###

*****************************************************************************/

static ERROR COMBOBOX_MoveToFront(struct ComboBox *Self, APTR Void)
{
   ActionMsg(AC_MoveToFront, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************
** ComboBox: NewObject()
*/

static ERROR COMBOBOX_NewObject(struct ComboBox *Self, APTR Void)
{
   OBJECTID var_id;
   OBJECTPTR variable;

   if (NewObject(ID_RENDER, NF_CHILD, NULL, &Self->RegionID) IS ERR_Okay) {
      if (NewObject(ID_FONT, NF_CHILD, &Self->Font, &Self->FontID) IS ERR_Okay) {
         if (NewObject(ID_TEXT, NF_CHILD, &Self->Text, &Self->TextID) IS ERR_Okay) {
            if (NewObject(ID_MENU, NF_CHILD, &Self->Menu, &Self->MenuID) IS ERR_Okay) {

               SetField(Self->Font, FID_Face, FT_STRING, "sans serif:9");

               Self->ReleaseFrame = 1;

               /*** Internal colour ***/
               Self->ColourRGB.Red   = 0;
               Self->ColourRGB.Green = 255;
               Self->ColourRGB.Blue  = 255;

               /*** Shadow colour ***/
               Self->ShadowRGB.Red   = 100;
               Self->ShadowRGB.Green = 100;
               Self->ShadowRGB.Blue  = 100;

               /*** Highlight colour ***/
               Self->HighlightRGB.Red   = 255;
               Self->HighlightRGB.Green = 255;
               Self->HighlightRGB.Blue  = 255;

               /*** Try to use colours from the user's preferences ***/

               if (FastFindObject("gl_colours", ID_VARIABLE, &var_id, 1, NULL) IS ERR_Okay) {
                  if (AccessObject(var_id, 4000, &variable) IS ERR_Okay) {
                     UBYTE buffer[100];

                     if (GetFieldVariable(variable, "background", buffer, sizeof(buffer)) IS ERR_Okay) {
                        SET_Colour(Self, buffer);
                     }

                     if (GetFieldVariable(variable, "highlight", buffer, sizeof(buffer)) IS ERR_Okay) {
                        SET_Highlight(Self, buffer);
                     }

                     if (GetFieldVariable(variable, "shadow", buffer, sizeof(buffer)) IS ERR_Okay) {
                        SET_Shadow(Self, buffer);
                     }

                     ReleaseObject(variable);
                  }
               }

               if (FastFindObject("gl_fonts", ID_VARIABLE, &var_id, 1, NULL) IS ERR_Okay) {
                  if (AccessObject(var_id, 4000, &variable) IS ERR_Okay) {
                     UBYTE buffer[100];

                     if (GetFieldVariable(variable, "window_face", buffer, sizeof(buffer)) IS ERR_Okay) {
                        SetField(Self->Font, FID_Face, FT_STRING, buffer);
                     }

                     if (GetFieldVariable(variable, "window_colour", buffer, sizeof(buffer)) IS ERR_Okay) {
                        SetField(Self->Font, FID_Colour, FT_STRING, buffer);
                     }

                     ReleaseObject(variable);
                  }
               }

               Self->Thickness = 1;
               Self->ButtonOffset = 1;

               SET_Template(Self, "templates:comboboxdef.xml");

               return(ERR_Okay);
            }
            else return(ERR_NewObject);
         }
         else return(ERR_NewObject);
      }
      else return(ERR_NewObject);
   }
   else return(ERR_NewObject);
}

/*****************************************************************************

###ACTION###
Name:  Redimension
Short: Changes the size and position of the combobox.
###END###

*****************************************************************************/

static ERROR COMBOBOX_Redimension(struct ComboBox *Self, struct acRedimension *Args)
{
   return(ActionMsg(AC_Redimension, Self->RegionID, Args));
}

/*****************************************************************************
** ComboBox: ReleaseObject
*/

static ERROR COMBOBOX_ReleaseObject(struct ComboBox *Self, APTR Void)
{
   if ((Self->Font) AND (Self->FontID)) {
      ReleaseObject(Self->Font);
      Self->Font = NULL;
   }

   if ((Self->Text) AND (Self->TextID)) {
      ReleaseObject(Self->Text);
      Self->Text = NULL;
   }

   if ((Self->Menu) AND (Self->MenuID)) {
      ReleaseObject(Self->Menu);
      Self->Menu = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Resize
Short: Alters the size of the combobox.
###END###

*****************************************************************************/

static ERROR COMBOBOX_Resize(struct ComboBox *Self, struct acResize *Args)
{
   return(ActionMsg(AC_Resize, Self->RegionID, Args));
}

/*****************************************************************************

###ACTION###
Name:  SetUnlistedField
Short: Arguments can be passed through to the combobox menu via unlisted fields.
###END###

*****************************************************************************/

static ERROR COMBOBOX_SetUnlistedField(struct ComboBox *Self, struct acSetUnlistedField *Args)
{
   Action(AC_SetUnlistedField, Self->Menu, Args);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Show
Short: Puts the combobox on display.
###END###

*****************************************************************************/

static ERROR COMBOBOX_Show(struct ComboBox *Self, APTR Void)
{
   ActionMsg(AC_Show, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ActionScript
Short:  Script to be executed when the combobox is clicked.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>When a combobox is activated, you can elect to run a script that
performs a responsive action.  The ActionScript string will typically refer
to a file location, although the special STRING: assignment followed with
the script execution text (e.g. DML code) can also be useful for optimising
the execution of simple scripts.</p>

###END###

*****************************************************************************/

static ERROR GET_ActionScript(struct ComboBox *Self, STRING *Value)
{
   if (Self->ActionScript) {
      *Value = Self->ActionScript;
      return(ERR_Okay);
   }
   else if (!Self->ActionScriptMID) {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
   else if (AccessMemory(Self->ActionScriptMID, MEM_READ, (void **)&Self->ActionScript) IS ERR_Okay) {
      *Value = Self->ActionScript;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ObjectError((OBJECTPTR)Self, ERH_GetField, ERR_AccessMemory));
   }
}

static ERROR SET_ActionScript(struct ComboBox *Self, STRING Value)
{
   LONG i;

   if (Self->ActionScript)    { ReleaseMemory(Self->ActionScript);   Self->ActionScript = NULL; }
   if (Self->ActionScriptMID) { FreeMemoryID(Self->ActionScriptMID); Self->ActionScriptMID = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR|Self->Head.MemFlags, (void **)&Self->ActionScript, &Self->ActionScriptMID) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->ActionScript[i] = Value[i];
         Self->ActionScript[i] = 0;
      }
      else return(ObjectError((OBJECTPTR)Self, ERH_SetField, ERR_AllocMemory));
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Border
Short:  String-based field for setting a single-colour border for the combobox.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The border colour for a combo box can be declared by writing to this field.
The colour must be in hexadecimal or separated-decimal format - for example
to create a pure red colour, a setting of "#ff0000" or "255,0,0" would be
valid.</p>

###SEE ALSO###
Fields: Highlight, Shadow, Thickness

###END###

*****************************************************************************/

static ERROR SET_Border(struct ComboBox *Self, STRING Colour)
{
   if (Colour) {
      StrReadColour(Colour, &Self->HighlightRGB.Red, &Self->HighlightRGB.Green, &Self->HighlightRGB.Blue, &Self->HighlightRGB.Alpha);
      Self->ShadowRGB.Red   = Self->HighlightRGB.Red;
      Self->ShadowRGB.Green = Self->HighlightRGB.Green;
      Self->ShadowRGB.Blue  = Self->HighlightRGB.Blue;
      Self->ShadowRGB.Alpha = Self->HighlightRGB.Alpha;
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Bottom
Short:  The bottom coordinate of the combobox.
Type:   LONG
Status: Get

###DESCRIPTION###
<p>The bottom coordinate of the combobox (calculated as YCoord + Height) is
readable from this field.</p>

###END###

*****************************************************************************/

static ERROR GET_Bottom(struct ComboBox *Self, LONG *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Bottom, FT_LONG, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
}

/*****************************************************************************

###FIELD###
Name:   ButtonOffset
Short:  Defines the button's pixel offset against the combobox borders.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>The positioning of the combobox button can be altered by setting
a value in this field.  Normally the button is offset from the combobox border
by 1 pixel, but depending on the combobox graphical setup, you may wish
to change this to a value of 0 or 2 for improved aesthetics.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Colour
Short:  String-based field for setting the colour inside of the combobox.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The colour inside of a combobox is declared by writing to this field.
The colour must be in hexadecimal or separated-decimal format - for example
to create a pure red colour, a setting of "#ff0000" or "255,0,0" would be
valid.</p>

###SEE ALSO###
Fields: Highlight, Shadow, Thickness

###END###

*****************************************************************************/

static ERROR SET_Colour(struct ComboBox *Self, STRING Colour)
{
   if (Colour) {
      StrReadColour(Colour, &Self->ColourRGB.Red, &Self->ColourRGB.Green, &Self->ColourRGB.Blue, &Self->ColourRGB.Alpha);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   FocusFrame
Short:  The graphics frame to display when the combobox has the focus.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>This field specifies the drawable frame to switch to when the user
focusses on the combobox.  By default this field is initially set to
zero, which has no effect on the drawable frame.  When the user leaves
the combobox, it will revert to the frame indicated by the ReleaseFrame
field.</p>

###SEE ALSO###
Field: ReleaseFrame

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   ColourRGB
Short:  Defines the internal colour of the combobox in RGB format.
Type:   RGB
Status: Read

###DESCRIPTION###
<p>The colour inside of the combobox can be read in RGB format from this field.</p>

###SEE ALSO###
Field: Colour

###END###

*****************************************************************************/

static ERROR GET_ColourRGB(struct ComboBox *Self, struct RGB *Value)
{
   Value->Red   = Self->ColourRGB.Red;
   Value->Green = Self->ColourRGB.Green;
   Value->Blue  = Self->ColourRGB.Blue;
   Value->Alpha = Self->ColourRGB.Alpha;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Disable
Short:  Disables the combobox on initalisation.
Type:   BOOLEAN
Status: Get/Set

###DESCRIPTION###
<p>The combobox can be disabled on initialisation by setting this field to TRUE.
If you need to disable the combobox after it has been activated, it is preferred
that you use the Disable action.</p>

<p>To enable the combobox after it has been disabled, use the Enable action.</p>

###END###

*****************************************************************************/

static ERROR GET_Disable(struct ComboBox *Self, LONG *Value)
{
   if (Self->Flags & CMF_DISABLED) *Value = TRUE;
   else *Value = FALSE;
   return(ERR_Okay);
}

static ERROR SET_Disable(struct ComboBox *Self, LONG *Value)
{
   if (*Value IS TRUE) Action(AC_Disable, Self, NULL);
   else if (*Value IS FALSE) Action(AC_Enable, Self, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Drawable
Short:  The drawable that will contain the combobox graphic.
Type:   OBJECTID
Status: Read/Init

###DESCRIPTION###
<p>The drawable that will contain the combobox graphic is set here.  If this
field is not set prior to initialisation, the combobox will attempt to scan
for the correct drawable by analysing its parents until it finds a suitable
candidate.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Flags
Short:  Optional flags may be set here.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>Optional flags that may be set against a combobox object are as follows:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">AUTOCOMPLETE</td><td>This option will assist the user when the combobox is in EDIT mode.  The combobox will attempt to help the user by automatically completing text as it is typed into the box.</td></tr>
<tr><td width="10%">DISABLED</td><td>This readable flag is automatically set when the combobox is disabled.  You may set it on initialisation if the combobox is to be disabled on creation.</td></tr>
<tr><td width="10%">EDIT</td><td>If you put the combobox into EDIT mode, the user will be able to type customised text into the combobox text area.</td></tr>
<tr><td width="10%">FOCUSACTIVATE</td><td>When in EDIT mode, if the combobox loses the focus, the presence of this option will force it to automtically activate itself (as if the user had pressed the enter key).</td></tr>
<tr><td width="10%">HIDE</td><td>This flag may be set on initialisation in order to hide the combobox on its creation (applies to script usage only).</td></tr>
<tr><td width="10%">LIMITTOLIST</td><td>If the combobox is in EDIT mode, the user will be able to type customised text data by default.  If you would prefer that the text is limited to that provided the combobox list, set this flag.</td></tr>
<tr><td width="10%">LOCAL</td><td>By default, the combobox will send messages to its child objects when the combobox value changes.  If you set the LOCAL option, the combobox will execute the objects within its own process space.</td></tr>
</table>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   FocusActivate
Short:  If TRUE, activation occurs when the user leaves the combobox area.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>If this field is set to TRUE, the combobox will automatically activate
itself if the user enters the combobox area and then leaves.  The activation
will occur regardless of wether or not the user edited the content of the
combobox.  The effect is the same as if the user had proactively pressed
the enter key.</p>

###END###
*****************************************************************************/

static ERROR GET_FocusActivate(struct ComboBox *Self, LONG *Value)
{
   if (Self->Flags & CMF_FOCUSACTIVATE) *Value = TRUE;
   else *Value = FALSE;
   return(ERR_Okay);
}

static ERROR SET_FocusActivate(struct ComboBox *Self, LONG *Value)
{
   if (*Value IS TRUE) Self->Flags |= CMF_FOCUSACTIVATE;
   else Self->Flags &= ~CMF_FOCUSACTIVATE;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Height
Short:  Defines the height of a combobox.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>An combobox can be given a fixed or relative height by setting this field
to the desired value.  To set a relative height, use the FD_PERCENT flag when
setting the field.</p>

###END###

*****************************************************************************/

static ERROR GET_Height(struct ComboBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetHeight, ERR_AccessObject));
}

static ERROR SET_Height(struct ComboBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (((Value->Type & FD_DOUBLE) AND (!Value->Double)) OR
       ((Value->Type & FD_LARGE) AND (!Value->Large))) {
      return(ERR_Okay);
   }

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_Height, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   Highlight
Short:  String-based field for setting the border highlight.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The border highlight of a combobox is set by writing to this field. The colour
must be in hexadecimal or separated-decimal format - for example to create a
pure red colour, a setting of "#ff0000" or "255,0,0" would be valid.</p>

###SEE ALSO###
Fields: Colour, Shadow, Thickness

###END###

*****************************************************************************/

static ERROR SET_Highlight(struct ComboBox *Self, STRING Value)
{
   if (Value) {
      StrReadColour(Value, &Self->HighlightRGB.Red, &Self->HighlightRGB.Green, &Self->HighlightRGB.Blue, &Self->HighlightRGB.Alpha);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   HighlightRGB
Short:  Defines border highlight of the combobox, in RGB format.
Type:   RGB
Status: Read

###DESCRIPTION###
<p>The border highlight of the combobox can be read in RGB format from this field.</p>

###SEE ALSO###
Field: Colour

###END###

*****************************************************************************/

static ERROR GET_HighlightRGB(struct ComboBox *Self, struct RGB *Value)
{
   Value->Red   = Self->HighlightRGB.Red;
   Value->Green = Self->HighlightRGB.Green;
   Value->Blue  = Self->HighlightRGB.Blue;
   Value->Alpha = Self->HighlightRGB.Alpha;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Label
Short:  The label is a string displayed to the left of the combobox area.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>A label can be drawn next to the combobox area by setting the Label field.
The label should be a short, descriptive string of one or two words.  It
is common practice for the label to be followed with a colon character.</p>

###END###

*****************************************************************************/

static ERROR GET_Label(struct ComboBox *Self, STRING *Value)
{
   *Value = Self->Label;
   return(ERR_Okay);
}

static ERROR SET_Label(struct ComboBox *Self, STRING Value)
{
   WORD i;
   if (Value) {
      for (i=0; (Value[i]) AND (i < sizeof(Self->Label)-1); i++) Self->Label[i] = Value[i];
      Self->Label[i] = 0;
   }
   else Self->Label[0] = 0;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   LabelWidth
Short:  A set-width for the label area of the combobox may be defined here.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>If you set a label for the combobox, the width of the label area is
automatically calculated according to the width of the label string.  You
may override this behaviour by setting a value in the LabelWidth field.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Region
Short:  The drawable that represents the combobox is referenced through this field.
Type:   OBJECTID
Status: Read

###DESCRIPTION###
<p>The rendering area that represents the combobox display can be accessed
through this field.  For further information, refer to the Render class.  Note
that talking to the drawable directly can have adverse effects on the combobox
control system.  Where possible, all communication should be limited to the
combobox object itself.</p>

###SEE ALSO###
Class: Render

###END###

*****************************************************************************/

static ERROR SET_Region(struct ComboBox *Self, LONG *Value)
{
   OBJECTPTR drawable;

   /* NOTE: For backwards compatibility with the Render class, the region
   ** can be set to a value of TRUE to define the combobox as a simple drawable
   ** region.
   */

   if ((*Value IS FALSE) OR (*Value IS TRUE)) {
      if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
         SetField(drawable, FID_Region, FT_LONG, Value);
         ReleaseObject(drawable);
         return(ERR_Okay);
      }
      else return(ERR_AccessObject);
   }
   else return(ERR_Failed);
}

/*****************************************************************************

###FIELD###
Name:   Menu
Short:  Provides direct access to the drop-down menu.
Type:   OBJECTPTR
Status: Read

###DESCRIPTION###
<p>The drop-down menu that is used for the combobox can be accessed directly
through this field.  You may find this useful for manipulating the content
of the drop-down menu following initialisation of the combobox.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   ReleaseFrame
Short:  The graphics frame to display when the combobox loses the focus.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>If the FocusFrame field has been set, you may want to match that value
by indicating the frame that should be used when the click is released.
By default, the value in this field will initially be set to 1.  This field
is unused if the FocusFrame field has not been set.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Right
Short:  The right-most coordinate of the combobox.
Type:   LONG
Status: Get

###DESCRIPTION###
<p>The right-most coordinate of the combobox (calculated as XCoord + Width) is
readable from this field.</p>

###END###

*****************************************************************************/

static ERROR GET_Right(struct ComboBox *Self, LONG *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Right, FT_LONG, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
}

/*****************************************************************************

###FIELD###
Name:   Shadow
Short:  String-based field for setting the combobox shadow.
Type:   STRING
Status: Set

###DESCRIPTION###
<p>The border shadow of a combobox is set by writing to this field. The colour
must be in hexadecimal or separated-decimal format - for example to create a
pure red colour, a setting of "#ff0000" or "255,0,0" would be valid.</p>

###SEE ALSO###
Fields: Colour, Highlight, Thickness

###END###

*****************************************************************************/

static ERROR SET_Shadow(struct ComboBox *Self, STRING Value)
{
   if (Value) {
      StrReadColour(Value, &Self->ShadowRGB.Red, &Self->ShadowRGB.Green, &Self->ShadowRGB.Blue, &Self->ShadowRGB.Alpha);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ShadowRGB
Short:  Defines the border shadow of the combobox, in RGB format.
Type:   RGB
Status: Read

###DESCRIPTION###
<p>The border shadow of the combobox can be read in RGB format from this field.</p>

###SEE ALSO###
Field: Colour

###END###

*****************************************************************************/

static ERROR GET_ShadowRGB(struct ComboBox *Self, struct RGB *Value)
{
   Value->Red   = Self->ShadowRGB.Red;
   Value->Green = Self->ShadowRGB.Green;
   Value->Blue  = Self->ShadowRGB.Blue;
   Value->Alpha = Self->ShadowRGB.Alpha;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   String
Short:  The string that is to be printed inside the combobox is declared here.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>The string that you would like to be displayed in the combobox is
specified in this field.  The string must be in UTF-8 format and may not
contain line feeds.  You can read this field at any time to determine what
the user has entered in the combobox.</p>

<p>If the string is changed after initialisation, the combobox will be redrawn
to show the updated text.</p>

###END###

*****************************************************************************/

static ERROR GET_String(struct ComboBox *Self, STRING *Value)
{
   STRING str;

   if (GetField(Self->Text, FID_String, FT_STRING, &str) IS ERR_Okay) {
      *Value = str;

      return(ERR_Okay);
   }
   else return(ERR_FieldNotSet);
}

static ERROR SET_String(struct ComboBox *Self, STRING Value)
{
   STRING original;

   /*** Do nothing if the string will remain unchanged ***/

   if ((GetField(Self->Text, FID_String, FT_STRING, &original) IS ERR_Okay) AND (original)) {
      if (StrCompare(original, Value, 0, STR_MATCHLENGTH) IS ERR_Okay) return(ERR_Okay);
   }

   if (SetField(Self->Text, FID_String, FT_STRING, Value) IS ERR_Okay) {
      if (Self->Head.ObjectFlags & NF_INITIALISED) {
         DPrintF("Set.String:","[ComboBox] Self-activating due to string setting.");
         Action(AC_Activate, Self, NULL);
      }
      return(ERR_Okay);
   }
   else return(ERR_Failed);
}

/*****************************************************************************

###FIELD###
Name:   TabFocus
Short:  Set this field to a TabFocus object to register the combobox in a tab-list.
Type:   OBJECTID
Status: Init

###DESCRIPTION###
<p>The TabFocus field provides a convenient way of adding the combobox to a
TabFocus object, so that it can be focussed on via the tab key.  Simply set
this field to the ID of the TabFocus object that is managing the tab-list
for the application window.</p>

###SEE ALSO###
Class: TabFocus

###END###

*****************************************************************************/

static ERROR SET_TabFocus(struct ComboBox *Self, OBJECTID *Value)
{
   OBJECTPTR tabfocus;

   if (AccessObject(*Value, 5000, &tabfocus) IS ERR_Okay) {
      if (tabfocus->ClassID IS ID_TABFOCUS) {
         ActionTags(MT_AddTabObject, tabfocus, Self->Head.UniqueID);
      }
      ReleaseObject(tabfocus);
   }
   else return(ERR_AccessObject);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Template
Short:  Defines the makeup of the combobox using a pre-defined template.
Type:   STRING
Status: Init

###DESCRIPTION###
<p>Setting the Template field allows a combobox defintion to be loaded from a
pre-defined template file.  The template may contain settings for all of the
field values of a combobox object, as well as a customised graphics script
for building the look and feel of the combobox.  The template must be provided
as an XML file that can contain two possible root tags - 'value' and 'graphics'.
The value tag contains a list of field values for setting against the combobox,
while the graphics tag contains the graphical makeup of the combobox.  The
following example illustrates how a customised combobox can be created:</p>

<pre>
&lt;?xml version="1.0"?&gt;

&lt;values&gt;
  &lt;thickness value="2"/&gt;
&lt;/values&gt;

&lt;graphics&gt;
  &lt;box boxes="(s0,0,!0,!0) (s1,1,!1,!1)" colour="230,230,230"
    highlight(0)="120,120,120" highlight(1)="255,255,255" shadow(0)="80,80,80"
    shadow(1)="200,200,200"/&gt;
&lt;/graphics&gt;
</pre>

###END###

*****************************************************************************/

static ERROR SET_Template(struct ComboBox *Self, STRING Value)
{
   struct XMLTag **xmllist, *tag, *child;
   UBYTE value[300];
   WORD i;

   if (Self->TemplateXML) {
      Action(AC_Free, Self->TemplateXML, NULL);
      Self->TemplateXML = NULL;
   }

   if (AnalyseLocation(Value, NULL) IS ERR_Okay) {
      if (CreateObject(ID_XML, NF_CHILD, &Self->TemplateXML, NULL,
         FID_Location|TSTRING, Value,
         TAGEND) IS ERR_Okay) {

         if (GetFields(Self->TemplateXML, FID_Tags|TPTR, &xmllist, TAGEND) IS ERR_Okay) {
            for (tag=xmllist[0]; tag; tag=tag->Next) {
               if (tag->Child) {
                  if (StrCompare("values", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                     for (child=tag->Child; child; child=child->Next) {
                        if (child->TotalAttributes > 1) {
                           for (i=0; (i < sizeof(value)-1) AND (child->Attributes[1].Value[i]); i++) value[i] = child->Attributes[1].Value[i];
                           value[i] = 0;
                           StrTranslate(value, sizeof(value), NULL);
                           SetFieldVariable(Self, child->Attributes[0].Name, value);
                        }
                     }
                     break;
                  }
               }
            }
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Thickness
Short:  The thickness of the combobox border.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>This field specifies the thickness of the combobox border, expressed in
pixels.  By default the thickness is set to a value of 1.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Width
Short:  Defines the width of a combobox.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>A combobox can be given a fixed or relative width by setting this field
to the desired value.  To set a relative width, use the FD_PERCENT flag when
setting the field.</p>

###END###

*****************************************************************************/

static ERROR GET_Width(struct ComboBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetWidth, ERR_AccessObject));
}

static ERROR SET_Width(struct ComboBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (((Value->Type & FD_DOUBLE) AND (!Value->Double)) OR
       ((Value->Type & FD_LARGE) AND (!Value->Large))) {
      return(ERR_Okay);
   }

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_Width, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   XCoord
Short:  The horizontal position of a combobox.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The horizontal position of a combobox can be set to an absolute or relative
coordinate by writing a value to the XCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Field: XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_X(struct ComboBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_XCoord, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetX, ERR_AccessObject));
}

static ERROR SET_X(struct ComboBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_XCoord, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   XOffset
Short:  The horizontal offset of a combobox.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The XOffset has a dual purpose depending on whether or not it is set in
conjunction with an X coordinate or a Width based field.</p>

<p>If set in conjunction with an X coordinate then the combobox will be
drawn from that X coordinate up to the width of the container, minus the
value given in the XOffset.  This means that the width of the ComboBox is
dynamically calculated in relation to the width of the container.</p>

<p>If the XOffset field is set in conjunction with a fixed or relative width
then the combobox will be drawn at an X coordinate calculated from the formula
&quot;XCoord = ContainerWidth - ComboBoxWidth - XOffset&quot;.</p>

###SEE ALSO###
Field: XCoord, Width, YOffset

###END###

*****************************************************************************/

static ERROR GET_XOffset(struct ComboBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_XOffset, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetXOffset, ERR_AccessObject));
}

static ERROR SET_XOffset(struct ComboBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_XOffset, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   YCoord
Short:  The vertical position of a combobox.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The vertical position of a ComboBox can be set to an absolute or relative
coordinate by writing a value to the YCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Field: YOffset, XCoord

###END###

*****************************************************************************/

static ERROR GET_Y(struct ComboBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_YCoord, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetY, ERR_AccessObject));

}

static ERROR SET_Y(struct ComboBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_YCoord, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   YOffset
Short:  The vertical offset of a combobox.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The YOffset has a dual purpose depending on whether or not it is set in
conjunction with a Y coordinate or a Height based field.</p>

<p>If set in conjunction with a Y coordinate then the combobox will be drawn
from that Y coordinate up to the height of the container, minus the value
given in the YOffset.  This means that the height of the combobox is
dynamically calculated in relation to the height of the container.</p>

<p>If the YOffset field is set in conjunction with a fixed or relative height
then the combobox will be drawn at a Y coordinate calculated from the formula
&quot;YCoord = ContainerHeight - ComboBoxHeight - YOffset&quot;.</p>

###SEE ALSO###
Field: YCoord, Height, XOffset

###END###

*****************************************************************************/

static ERROR GET_YOffset(struct ComboBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_YOffset, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetYOffset, ERR_AccessObject));
}

static ERROR SET_YOffset(struct ComboBox *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_YOffset, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************
** Internal: DrawComboBox()
*/

static void DrawComboBox(struct ComboBox *Self, struct Render *Drawable, struct Bitmap *Bitmap)
{
   struct mtDrawRGBRectangle rect;
   struct Font *font;
   ULONG highlight, shadow, temp;
   WORD i, x, width;

   if (!(Self->Flags & CMF_NOBKGD)) {
      rect.XCoord = Self->LabelWidth;
      rect.YCoord = 0;
      rect.Width  = Drawable->Width - Self->LabelWidth;
      rect.Height = Drawable->Height;
      rect.Fill   = TRUE;
      rect.Red    = Self->ColourRGB.Red;
      rect.Green  = Self->ColourRGB.Green;
      rect.Blue   = Self->ColourRGB.Blue;
      Action(MT_DrawRGBRectangle, Bitmap, &rect);

      /*** Draw the borders around the rectangular area ***/

      shadow    = PackPixel(Bitmap, Self->ShadowRGB.Red, Self->ShadowRGB.Green, Self->ShadowRGB.Blue);
      highlight = PackPixel(Bitmap, Self->HighlightRGB.Red, Self->HighlightRGB.Green, Self->HighlightRGB.Blue);

      /*** Reverse the border definitions in sunken mode ***/
      temp      = highlight;
      highlight = shadow;
      shadow    = temp;

      x = Self->LabelWidth;
      width = Drawable->Width - Self->LabelWidth;
      if (Self->Thickness > 0) {
         for (i=0; i < Self->Thickness; i++) {
            /*** Top, Bottom ***/

            ActionTags(MT_DrawRectangle, Bitmap, x+i, i, width-i-i, 1, highlight, TRUE);
            ActionTags(MT_DrawRectangle, Bitmap, x+i, Drawable->Height-i-1, width-i-i, 1, shadow, TRUE);

            /*** Left, Right ***/

            ActionTags(MT_DrawRectangle, Bitmap, x+i,         i+1, 1, Drawable->Height-i-i-2, highlight, TRUE);
            ActionTags(MT_DrawRectangle, Bitmap, x+width-i-1, i+1, 1, Drawable->Height-i-i-2, shadow, TRUE);
         }
      }
   }

   if (Self->Label[0]) {
      font = Self->Font;

      SetField(font, FID_Bitmap, FT_POINTER, Bitmap);
      SetField(font, FID_String, FT_STRING, Self->Label);

      if (Drawable->Flags & RNF_DISABLED) SetFields(font, FID_Opacity|TLONG, 25, TAGEND);

      font->XCoord = 0;
      font->YCoord = 0;
      font->AlignFlags |= ALIGN_VERTICAL;
      font->AlignWidth  = Drawable->Width;
      font->AlignHeight = Drawable->Height;
      Action(AC_Draw, font, NULL);

      if (Drawable->Flags & RNF_DISABLED) SetFields(font, FID_Opacity|TLONG, 0, TAGEND);
   }
}








