/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      ScrollBar
Version:   1.0
ID:        ID_SCROLLBAR
Status:    Beta
Category:  GUI
Date:      December 2003
Author:    Rocklyte Systems
Copyright: Rocklyte Systems, 2003.  All rights reserved.
Keywords:  scrollbar, scroll, bar, up, down, left, right
Short:     The ScrollBar class creates scroll-bars for the user interface.

###DESCRIPTION###
<p>The ScrollBar class simplifies the creation and management of scrollbars
as part of the user interface.  A scrollbar is a rectangular gadget that
typically sits at the far-right or bottom of a scrollable page.  A slider is
present in the scrollbar and performs the scrolling action when it is
dragged with the mouse.  Buttons may be present at either end of the scrollbar
for the purpose of micro-scrolling.</p>

<p>The definitions for new scrollbars are loaded by default from the
environment file "templates:scrollbardef.xml".  You can change the template
file prior to initialisation by setting the Template field.  For information
on how to develop customised scrollbars, please refer to the documentation
for the Template field.</p>

<p>The ScrollBar class is closely related to the Scroll class.  To configure
the size of the scrollable page and the viewable area, you need to
communicate that information to the Scroll class.  More information is available
in the Scroll class documentation.</p>

<p>To create a new scrollbar, you need to specify the scrollbar's
scrolling direction at a minimum.  The following example illustrates:</p>

<pre>
  &lt;scrollbar direction="horizontal"/&gt;
</pre>

<p>The positioning of the scrollbar will be automatically calculated for you
based on this information.  To link a scrollbar to an object, such as a text
viewing area, you will need to extract the scroll object from the scrollbar
and pass it to the scrollable object.  Here is an example:</p>

<pre>
  &lt;scrollbar name="vsb" direction="vertical"/&gt;
  &lt;text vscroll="[vsb.scroll]" .../&gt;
</pre>

<p>If a new scrollbar is created without being hooked into another object,
it will send scroll messages to its parent drawable.</p>

###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/files/all.h>
#include <pandora/misc/all.h>
#include <pandora/data/xml.h>
#include <pandora/functions/set.c>
#include <pandora/functions/inttostr.c>
#include <pandora/functions/strtoint.c>
#include <pandora/functions/strcopy.c>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_SCROLLBAR  1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_SCROLLBAR, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2003. All rights reserved.",
   "December 2003",
   "ScrollBar"
};

static struct KernelBase  *KernelBase;
static OBJECTPTR ScrollBarClass = NULL;
static OBJECTPTR ScrollbarMod = NULL;
static OBJECTPTR glTemplateScript = NULL;

/*****************************************************************************
** Class definition.
*/

static struct FieldDef ScrollBarFlags[] = {
   { "HIDE",        SBF_HIDE },
   { "CONSTANT",    SBF_CONSTANT },
   { "RELATIVE",    SBF_RELATIVE },
   { "SLIDER",      SBF_SLIDER },
   { "NOINTERSECT", SBF_NOINTERSECT },
   { NULL, NULL }
};

static struct FieldDef DirectionLookup[] = {
   { "HORIZONTAL", SO_HORIZONTAL },
   { "VERTICAL",   SO_VERTICAL },
   { NULL, NULL }
};

static ERROR GET_Bottom(struct ScrollBar *, LONG *);
static ERROR GET_Height(struct ScrollBar *, struct Variable *);
static ERROR GET_Right(struct ScrollBar *, LONG *);
static ERROR GET_Template(struct ScrollBar *, STRING *);
static ERROR GET_Width(struct ScrollBar *, struct Variable *);
static ERROR GET_X(struct ScrollBar *, struct Variable *);
static ERROR GET_XOffset(struct ScrollBar *, struct Variable *);
static ERROR GET_Y(struct ScrollBar *, struct Variable *);
static ERROR GET_YOffset(struct ScrollBar *, struct Variable *);

static ERROR SET_Height(struct ScrollBar *, struct Variable *);
static ERROR SET_Hide(struct ScrollBar *, LONG *);
static ERROR SET_Template(struct ScrollBar *, STRING);
static ERROR SET_Width(struct ScrollBar *, struct Variable *);
static ERROR SET_X(struct ScrollBar *, struct Variable *);
static ERROR SET_XOffset(struct ScrollBar *, struct Variable *);
static ERROR SET_Y(struct ScrollBar *, struct Variable *);
static ERROR SET_YOffset(struct ScrollBar *, struct Variable *);

static struct FieldArray ScrollBarFields[] = {
   { "Region",    0, FDF_OBJECTID|FDF_RW,  ID_RENDER, NULL, NULL },
   { "Drawable",  0, FDF_OBJECTID|FDF_RW,  ID_RENDER, NULL, NULL },
   { "Slider",    0, FDF_OBJECTID|FDF_RW,  ID_RENDER, NULL, NULL },
   { "Flags",     0, FDF_LONGFLAGS|FDF_RW, (LONG)&ScrollBarFlags, NULL, NULL },
   { "Scroll",    0, FDF_CHILD|FDF_R,      ID_SCROLL, NULL, NULL },
   { "Direction", 0, FDF_LONG|FDF_LOOKUP|FDF_RI, (LONG)&DirectionLookup, NULL, NULL },
   { "Opacity",   0, FDF_FLOAT|FDF_RI,     0, NULL, NULL },
   { "Breadth",   0, FDF_LONG|FDF_RI,      0, NULL, NULL },
   { "Intersect", 0, FDF_OBJECTID|FDF_RI,  ID_SCROLLBAR, NULL, NULL },
   /** Virtual fields ***/
   { "Bottom",   0, FDF_LONG|FDF_R,      0, GET_Bottom,   NULL },
   { "Right",    0, FDF_LONG|FDF_R,      0, GET_Right,    NULL },
   { "Template", 0, FDF_STRING|FDF_RI,   0, GET_Template, SET_Template },
   { "Hide",     0, FDF_LONG|FDF_I,     0, NULL, SET_Hide },
   /*** Variable Fields ***/
   { "Height",   0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Height,  SET_Height },
   { "Width",    0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Width,   SET_Width },
   { "X",        0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XCoord",   0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XOffset",  0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_XOffset, SET_XOffset },
   { "Y",        0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YCoord",   0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YOffset",  0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_YOffset, SET_YOffset },
   END_FIELD
};

/*****************************************************************************
** Actions
*/

static ERROR SCROLLBAR_AccessObject(struct ScrollBar *, APTR);
static ERROR SCROLLBAR_ActionNotify(struct ScrollBar *, struct acActionNotify *);
static ERROR SCROLLBAR_ClosingTag(struct ScrollBar *, APTR);
static ERROR SCROLLBAR_Free(struct ScrollBar *, APTR);
static ERROR SCROLLBAR_Hide(struct ScrollBar *, APTR);
static ERROR SCROLLBAR_Init(struct ScrollBar *, APTR);
static ERROR SCROLLBAR_NewObject(struct ScrollBar *, APTR);
static ERROR SCROLLBAR_Redimension(struct ScrollBar *, struct acRedimension *);
static ERROR SCROLLBAR_ReleaseObject(struct ScrollBar *, APTR);
static ERROR SCROLLBAR_Resize(struct ScrollBar *, struct acResize *);
static ERROR SCROLLBAR_Show(struct ScrollBar *, APTR);

static struct ActionArray ScrollBarActions[] = {
   { AC_AccessObject,  SCROLLBAR_AccessObject },
   { AC_ActionNotify,  SCROLLBAR_ActionNotify },
   { AC_ClosingTag,    SCROLLBAR_ClosingTag },
   { AC_Free,          SCROLLBAR_Free },
   { AC_Hide,          SCROLLBAR_Hide },
   { AC_Init,          SCROLLBAR_Init },
   { AC_NewObject,     SCROLLBAR_NewObject },
   { AC_Redimension,   SCROLLBAR_Redimension },
   { AC_ReleaseObject, SCROLLBAR_ReleaseObject },
   { AC_Resize,        SCROLLBAR_Resize },
   { AC_Show,          SCROLLBAR_Show },
   { NULL, NULL }
};

/*****************************************************************************
** Scrollbar methods.
*/

static struct MethodArray ScrollBarMethods[] = {
   { NULL, NULL, NULL, NULL, NULL }
};

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   /*** Load the default template file ***/

   if (CreateObject(ID_SCRIPT, NF_CHILD, &glTemplateScript, NULL,
      FID_Location|TSTRING, "templates:scrollbardef.xml",
      TAGEND) IS ERR_Okay) {
   }

   if (GetField(argModule, FID_Master, FT_POINTER, &ScrollbarMod) IS ERR_Okay) {
   }

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&ScrollBarClass, NULL,
      FID_BaseClassID|TLONG, ID_SCROLLBAR,
      FID_Version|TFLOAT,    VER_SCROLLBAR,
      FID_Name|TSTRING,      "ScrollBar",
      FID_Category|TLONG,    CCF_GUI,
      FID_Flags|TLONG,       CLF_PROMOTECHILDREN,
      FID_Actions|TPTR,      ScrollBarActions,
      FID_Methods|TPTR,      ScrollBarMethods,
      FID_Fields|TPTR,       ScrollBarFields,
      FID_Size|TLONG,        sizeof(struct ScrollBar),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (glTemplateScript) { acFree(glTemplateScript); glTemplateScript = NULL; }
   if (ScrollBarClass)   { acFree(ScrollBarClass);   ScrollBarClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** ScrollBar: AccessObject
*/

static ERROR SCROLLBAR_AccessObject(struct ScrollBar *Self, APTR Void)
{
   if (Self->ScrollID) {
      if (AccessObject(Self->ScrollID, 0, (OBJECTPTR *)&Self->Scroll) != ERR_Okay) {
         return(ObjectError(Self, ERH_AccessObject, ERR_AccessObject));
      }
   }
   return(ERR_Okay);
}

/*****************************************************************************
** ScrollBar: ActionNotify
*/

static ERROR SCROLLBAR_ActionNotify(struct ScrollBar *Self, struct acActionNotify *Args)
{
   return(ERR_Okay);
}

/*****************************************************************************
** ScrollBar: ClosingTag
*/

static ERROR SCROLLBAR_ClosingTag(struct ScrollBar *Self, APTR Void)
{
   if (Self->Flags & SBF_CONSTANT) Action(AC_Show, Self, NULL);
   return(ERR_Okay);
}

/*****************************************************************************
** ScrollBar: Free
*/

static ERROR SCROLLBAR_Free(struct ScrollBar *Self, APTR Void)
{
   if (Self->ScrollID) {
      if (Self->Scroll) {
         acFree(Self->Scroll);
         ReleaseObject(Self->Scroll);
         Self->Scroll = NULL;
      }
      else acFreeID(Self->ScrollID);
      Self->ScrollID = NULL;
   }

   if (Self->RegionID) {
      acFreeID(Self->RegionID);
      Self->RegionID = NULL;
   }

   if (Self->Script) {
      acFree(Self->Script);
      Self->Script = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Hide
Short: Removes the scrollbar from the display.
###END###

*****************************************************************************/

static ERROR SCROLLBAR_Hide(struct ScrollBar *Self, APTR Void)
{
   acHideID(Self->RegionID);
   return(ERR_Okay);
}

/*****************************************************************************
** ScrollBar: Init()
*/

static ERROR SCROLLBAR_Init(struct ScrollBar *Self, APTR Void)
{
   struct Render *region, *slider;
   struct ScrollBar *intersect;
   OBJECTPTR script, file;
   OBJECTID owner_id, monitorid, objectid;
   LONG flags, count;
   WORD i;

   /*** Find our parent drawable ***/

   if (!Self->DrawableID) {
      owner_id = GetOwner(Self);
      while ((owner_id) AND (GetClassID(owner_id) != ID_RENDER)) {
         owner_id = GetOwnerID(owner_id);
      }
      if (owner_id) Self->DrawableID = owner_id;
      else return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
   }

   if ((glTemplateScript) AND (StrCompare("templates:scrollbardef.xml", Self->Template, 0, STR_MATCHLENGTH) IS ERR_Okay)) {
      /* Check if the template script has changed since the last time we loaded
      ** (typically occurs if the user has switched environments).
      */
      if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
         FID_Location|TSTRING, "templates:scrollbardef.xml",
         TAGEND) IS ERR_Okay) {
         LARGE file_timestamp, file_filesize, script_timestamp, script_filesize;

         GetFields(file, FID_TimeStamp|TLARGE, &file_timestamp,
                         FID_Size|TLARGE,  &file_filesize,
                         TAGEND);

         Action(AC_Free, file, NULL);

         GetFields(glTemplateScript, FID_TimeStamp|TLARGE, &script_timestamp,
                                     FID_FileSize|TLARGE,  &script_filesize,
                                     TAGEND);

         if ((file_timestamp != script_timestamp) OR (file_filesize != script_filesize)) {
            DPrintF("Init:","[ScrollBar] Reloading the template script.");
            if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
               FID_Location|TSTRING, "templates:scrollbardef.xml",
               TAGEND) IS ERR_Okay) {

               SetOwner(script, ScrollbarMod);

               Action(AC_Free, glTemplateScript, NULL); /* Free the existing template */
               glTemplateScript = script; /* Replace it with the new one */
            }
         }

      }
      script = glTemplateScript;
   }
   else if (CreateObject(ID_SCRIPT, NF_CHILD, &Self->Script, NULL,
      FID_Location|TSTRING, Self->Template,
      TAGEND) IS ERR_Okay) {
      script = Self->Script;
   }
   else return(ObjectError(Self, ERH_Init, ERR_CreateObject));

   UBYTE buffer[30];
   buffer[0] = '#';
   IntToStr(Self->ScrollID, buffer+1, sizeof(buffer)-1);

   SetUnlistedField(script, "scroll", buffer);

   /*** Initialise the main scrollbar region ***/

   if (AccessObject(Self->RegionID, 5000, (OBJECTPTR *)&region) IS ERR_Okay) {
      SetFields(region, FID_Parent|TLONG, Self->DrawableID,
                        FID_Opacity|TFLOAT, Self->Opacity,
                        TAGEND);

      if (!Self->Direction) {
         BYTE count;
         /* Try to determine the scrollbar direction based on the dimensions
         ** that have already been set.
         */
         count = 0;
         if (region->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) count++;
         if (region->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET)) count++;
         if (region->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT)) count++;
         if (count > 1) Self->Direction = SO_VERTICAL;
         else {
            count = 0;
            if (region->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) count++;
            if (region->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET)) count++;
            if (region->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH)) count++;
            if (count > 1) Self->Direction = SO_HORIZONTAL;
         }
      }
                        
      if (Self->Direction IS SO_HORIZONTAL) {
         if (!(region->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X))) SetFields(region, FID_XCoord|TLONG, 0, TAGEND);

         if ((!(region->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH))) AND (!(region->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET)))) {
            SetFields(region, FID_XOffset|TLONG, 0, TAGEND);
         }

         if ((!(region->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y))) AND (!(region->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET)))) {
            SetFields(region, FID_YOffset|TLONG, 0, TAGEND);
         }

         if (!(region->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) SetFields(region, FID_Height|TLONG, Self->Breadth, TAGEND);
      }

      if (Self->Direction IS SO_VERTICAL) {
         if (!(region->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y))) SetFields(region, FID_YCoord|TLONG, 0, TAGEND);

         if ((!(region->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) AND (!(region->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET)))) {
            SetFields(region, FID_YOffset|TLONG, 0, TAGEND);
         }

         if ((!(region->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X))) AND (!(region->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET)))) {
            SetFields(region, FID_XOffset|TLONG, 0, TAGEND);
         }

         if (!(region->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH))) SetFields(region, FID_Width|TLONG, Self->Breadth, TAGEND);
      }

      if (acInit(region) IS ERR_Okay) {

         if (!Self->Direction) {
            if (region->Width > region->Height) Self->Direction = SO_HORIZONTAL;
            else Self->Direction = SO_VERTICAL;
         }

         /*** Run the graphical script for the scrollbar area ***/

         if (Self->Direction IS SO_HORIZONTAL) SetField(script, FID_Procedure, FT_STRING, "hbar");
         else SetField(script, FID_Procedure, FT_STRING, "vbar");
         SetField(script, FID_Target, FT_LONG, &Self->RegionID);
         acActivate(script);

         /*** Initialise the slider region ***/

         if (AccessObject(Self->SliderID, 5000, (OBJECTPTR *)&slider) IS ERR_Okay) {
            STRING movement;
            SetOwner(slider, region);

            if (Self->Direction IS SO_HORIZONTAL) movement = "horizontal";
            else movement = "vertical";

            SetFields(slider, FID_XCoord|TLONG,      region->LeftMargin,
                              FID_YCoord|TLONG,      region->TopMargin,
                              FID_Width|TLONG,       region->Width - region->LeftMargin - region->RightMargin,
                              FID_Height|TLONG,      region->Height - region->TopMargin - region->BottomMargin,
                              FID_Drag|TLONG,        Self->SliderID,
                              FID_TopLimit|TLONG,    region->TopMargin,
                              FID_LeftLimit|TLONG,   region->LeftMargin,
                              FID_RightLimit|TLONG,  region->RightMargin,
                              FID_BottomLimit|TLONG, region->BottomMargin,
                              FID_Movement|TSTRING,  movement,
                              FID_Region|TLONG,      TRUE,
                              TAGEND);

            if (acInit(slider) IS ERR_Okay) {

               /*** Run the graphical script for the slider ***/

               if (Self->Direction IS SO_HORIZONTAL) SetField(script, FID_Procedure, FT_STRING, "hslider");
               else SetField(script, FID_Procedure, FT_STRING, "vslider");
               SetField(script, FID_Target, FT_LONG, &Self->SliderID);
               acActivate(script);
            }

            acShow(slider);

            ReleaseObject(slider);
         }
         else return(ObjectError(Self, ERH_Init, ERR_AccessObject));
      }
      else {
         ReleaseObject(region);
         return(ObjectError(Self, ERH_Init, ERR_Init));
      }

      ReleaseObject(region);
   }
   else return(ObjectError(Self, ERH_Init, ERR_AccessObject));

   /* If no intersecting scrollbar has been specified, check our parent drawable to see
   ** if an opposed scrollbar exists.
   */

   if ((!Self->IntersectID) AND (!(Self->Flags & SBF_NOINTERSECT))) {
      if ((count = TotalChildren(Self->DrawableID))) {
         struct ChildEntry list[count];
         if (ListChildren(Self->DrawableID, list, &count) IS ERR_Okay) {
            for (i=0; i < count; i++) {
               if ((list[i].ClassID IS ID_SCROLLBAR) AND (list[i].ObjectID != Self->Head.UniqueID)) {
                  if (AccessObject(list[i].ObjectID, 5000, (OBJECTPTR *)&intersect) IS ERR_Okay) {
                     if ((intersect->Direction IS SO_HORIZONTAL) AND (Self->Direction IS SO_VERTICAL)) Self->IntersectID = list[i].ObjectID;
                     else if ((intersect->Direction IS SO_VERTICAL) AND (Self->Direction IS SO_HORIZONTAL)) Self->IntersectID = list[i].ObjectID;
                     ReleaseObject(intersect);
                  }
                  break;
               }
            }
         }
      }
   }

   /*** Initialise the scroll management object ***/

   flags = NULL;
   if (!(Self->Flags & SBF_CONSTANT)) flags |= SCF_AUTOHIDE;
   if (Self->Flags & SBF_RELATIVE) flags |= SCF_RELATIVE;
   if (Self->Flags & SBF_SLIDER) flags |= SCF_SLIDER;
   if (Self->Direction IS SO_HORIZONTAL) flags |= SCF_HORIZONTAL;
   else flags |= SCF_VERTICAL;

   if (Self->IntersectID) {
      if (GetClassID(Self->IntersectID) IS ID_SCROLLBAR) {
         if (AccessObject(Self->IntersectID, 5000, (OBJECTPTR *)&intersect) IS ERR_Okay) {
            Self->IntersectID = intersect->ScrollID;
            ReleaseObject(intersect);
         }
      }
   }

   /*** If the Scroll.Object field has not been set, set it to our parent drawable ***/

   if ((GetField(Self->Scroll, FID_Object, FT_LONG, &objectid) != ERR_Okay) OR (!objectid)) {
      SetField(Self->Scroll, FID_Object, FT_LONG, &Self->DrawableID);
      objectid = Self->DrawableID;
   }

   /*** If the Scroll.Monitor field is not set, set it to the parent drawable ***/

   if ((GetField(Self->Scroll, FID_Monitor, FT_LONG, &monitorid) != ERR_Okay) OR (!monitorid)) {
      if (GetClassID(objectid) IS ID_RENDER) SetField(Self->Scroll, FID_Monitor, FT_LONG, &objectid);
      else SetField(Self->Scroll, FID_Monitor, FT_LONG, &Self->DrawableID);
   }

   if (SetFields(Self->Scroll, FID_Slider|TLONG,    Self->SliderID,
                               FID_Flags|TLONG,     flags,
                               FID_Intersect|TLONG, Self->IntersectID,
                               TAGEND) IS ERR_Okay) {
      UBYTE name[40];
      GetName(Self, name, sizeof(name));
      for (i=0; name[i]; i++);
      StrCopy("_scroll", name+i, sizeof(name)-i);

      SetName(Self->Scroll, name);
      if (acInit(Self->Scroll) IS ERR_Okay) {

      }
   }

   /*** Create the buttons for the scrollbar ***/

   if (Self->Direction IS SO_HORIZONTAL) SetField(script, FID_Procedure, FT_STRING, "hbuttons");
   else SetField(script, FID_Procedure, FT_STRING, "vbuttons");
   SetField(script, FID_Target, FT_LONG, &Self->RegionID);
   acActivate(script);

   if (Self->Script) {
      acFree(Self->Script);
      Self->Script = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** ScrollBar: NewObject()
*/

static ERROR SCROLLBAR_NewObject(struct ScrollBar *Self, APTR Void)
{
   OBJECTPTR slider, var;
   OBJECTID varid;

   if (NewObject(ID_RENDER, NF_CHILD, NULL, &Self->RegionID) IS ERR_Okay) {
      if (NewObject(ID_RENDER, NULL, &slider, &Self->SliderID) IS ERR_Okay) {
         if (NewObject(ID_SCROLL, Self->Head.ObjectFlags, &Self->Scroll, &Self->ScrollID) IS ERR_Okay) {
            SetOwner(Self->Scroll, slider);
            SET_Template(Self, "templates:scrollbardef.xml");
            ReleaseObject(slider);

            Self->Breadth = 16;
            Self->Opacity = 100;

            if (FastFindObject("gl_interface", ID_VARIABLE, &varid, 1, NULL) IS ERR_Okay) {
               if (AccessObject(varid, 4000, &var) IS ERR_Okay) {
                  UBYTE buffer[100];

                  if (GetFieldVariable(var, "slidersize", buffer, sizeof(buffer)) IS ERR_Okay) {
                     Self->Breadth = StrToInt(buffer);
                  }

                  if (GetFieldVariable(var, "scrollopacity", buffer, sizeof(buffer)) IS ERR_Okay) {
                     Self->Opacity = StrToInt(buffer);
                     if (Self->Opacity < 50) Self->Opacity = 50;
                  }

                  ReleaseObject(var);
               }
            }

            return(ERR_Okay);
         }
         else {
            ReleaseObject(slider);
            return(ERR_NewObject);
         }
      }
      else return(ERR_NewObject);
   }
   else return(ERR_NewObject);
}

/*****************************************************************************

###ACTION###
Name:  Redimension
Short: Changes the size and position of the scrollbar.
###END###

*****************************************************************************/

static ERROR SCROLLBAR_Redimension(struct ScrollBar *Self, struct acRedimension *Args)
{
   return(ActionMsg(AC_Redimension, Self->RegionID, Args));
}

/*****************************************************************************
** ScrollBar: ReleaseObject
*/

static ERROR SCROLLBAR_ReleaseObject(struct ScrollBar *Self, APTR Void)
{
   if ((Self->Scroll) AND (Self->ScrollID)) {
      ReleaseObject(Self->Scroll);
      Self->Scroll = NULL;
   }
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Resize
Short: Alters the size of the scrollbar.
###END###

*****************************************************************************/

static ERROR SCROLLBAR_Resize(struct ScrollBar *Self, struct acResize *Args)
{
   return(ActionMsg(AC_Resize, Self->RegionID, Args));
}

/*****************************************************************************

###ACTION###
Name:  Show
Short: Puts the scrollbar on display.
###END###

*****************************************************************************/

static ERROR SCROLLBAR_Show(struct ScrollBar *Self, APTR Void)
{
   ActionMsg(AC_Show, Self->RegionID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Bottom
Short:  The bottom coordinate of the scrollbar.
Type:   LONG
Status: Get

###DESCRIPTION###
<p>The bottom coordinate of the scrollbar (calculated as YCoord + Height) is
readable from this field.</p>

###END###

*****************************************************************************/

static ERROR GET_Bottom(struct ScrollBar *Self, LONG *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Bottom, FT_LONG, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
}

/*****************************************************************************

###FIELD###
Name:   Breadth
Short:  The breadth of the scrollbar is defined here.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>The breadth of the scrollbar can be altered in this field.  The breadth
of the scrollbar will either be its width (if vertical) or height (if
horizontal). The breadth of the scrollbar is user-definable and should only
be set with good reason.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Drawable
Short:  The drawable that will contain the scrollbar graphic.
Type:   OBJECTID
Status: Read/Init

###DESCRIPTION###
<p>The drawable that will contain the scrollbar graphic is set here.  If this
field is not set prior to initialisation, the scrollbar will attempt to scan
for the correct drawable by analysing its parents until it finds a suitable
candidate.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Flags
Short:  Optional flags may be set here.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>Optional flags that may be set against a scrollbar object are as follows:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">HIDE</td><td>This flag may be set on initialisation in order to hide the scrollbar on its creation (applies to script usage only).</td></tr>
<tr><td width="10%">CONSTANT</td><td>By default, a scrollbar will automatically remove itself from the display when the area to be scrolled is smaller than the size of the view.  You can turn off this behaviour by setting the CONSTANT option, which ensures that the scrollbar is always on display.</td></tr>
<tr><td width="10%">NOINTERSECT</td><td>If you create two scrollbars in the same drawable space, they will automatically be configured to intersect with each other.  You can turn off this behaviour by setting the NOINTERSECT flag.</td></tr>
</table>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Height
Short:  Defines the height of a scrollbar.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>A scrollbar can be given a fixed or relative height by setting this field
to the desired value.  To set a relative height, use the FD_PERCENT flag when
setting the field.</p>

###END###

*****************************************************************************/

static ERROR GET_Height(struct ScrollBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetHeight, ERR_AccessObject));
}

static ERROR SET_Height(struct ScrollBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (((Value->Type & FD_DOUBLE) AND (!Value->Double)) OR
       ((Value->Type & FD_LARGE) AND (!Value->Large))) {
      return(ERR_Okay);
   }

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_Height, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

static ERROR SET_Hide(struct ScrollBar *Self, LONG *Value)
{
   if (*Value IS TRUE) Self->Flags |= SBF_HIDE;
   else Self->Flags &= ~SBF_HIDE;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Right
Short:  The right coordinate of the scrollbar.
Type:   LONG
Status: Get

###DESCRIPTION###
<p>The right coordinate of the scrollbar (calculated as XCoord + Width) is
readable from this field.</p>

###END###

*****************************************************************************/

static ERROR GET_Right(struct ScrollBar *Self, LONG *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Right, FT_LONG, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
}

/*****************************************************************************

###FIELD###
Name:   Template
Short:  Defines the makeup of the scrollbar using a pre-defined template.
Type:   STRING
Status: Init

###DESCRIPTION###
<p>Customised scrollbars can be created by writing a template file and
then referring to its location in this field.  You will be required to
configure the look of the scrollbar for both vertical and horizontal
arrangements.  This is done by writing a series of script procedures for
the different areas of the scrollbar.</p>

<p>The following extract is from the default templates:scrollbardef.xml
file.  Notice that there are three areas to define, all of which relate to
the creation of a vertical scrollbar.  The procedures are: vbar - defines
the background graphic for the entire scrollbar area; vslider - defines the
graphical content of the slider; vbuttons - creates the buttons for the
scrollbar.  For a horizontal scrollbar the procedures are named hbar,
hslider and hbuttons.  A special 'scroll' argument may be referenced if you
need to talk to the scroll object that controls the scrollbar slider.  This
is necessary for registering buttons in the scrollbar (note use of the Scroll class'
AddButton method).</p>

<pre>
&lt;dml procedure="vbar"&gt;
  &lt;set object="[owner]" width="18" topmargin="=[owner.width]-1" bottommargin="=[owner.width]-1"/&gt;
  &lt;gradient firstcolour="192,192,192" lastcolour="90,90,90" direction="vertical"/&gt;
  &lt;box border="64,64,64"/&gt;
&lt;/dml&gt;

&lt;dml procedure="vslider"&gt;
  &lt;box boxes="(r0,0,!0,!0) (r1,1,!1,!1)" highlight="64,64,64"
    highlight(1)="255,255,255" shadow="64,64,64" shadow(1)="175,175,175" colour="230,230,230"/&gt;
  &lt;image src="environment:images/scrollbar" align="center"/&gt;
&lt;/dml&gt;

&lt;dml procedure="vbuttons"&gt;
  &lt;button x="0" y="0" height="[owner.width]" xoffset="0" flags="!nofocus"&gt;
    &lt;arrow drawable="[owner.region]" align="center" arrowwidth="7" arrowheight="4" direction="up" colour="60,60,60"/&gt;
    &lt;action method="addbutton" object="[{scroll}]" &button="[owner]" &direction="1"/&gt;
  &lt;/button&gt;

  &lt;button x="0" yoffset="0" height="[owner.width]" xoffset="0" flags="!nofocus"&gt;
    &lt;arrow drawable="[owner.region]" align="center" arrowwidth="7" arrowheight="4" direction="down" colour="60,60,60"/&gt;
    &lt;action method="addbutton" object="[{scroll}]" &button="[owner]" &direction="2"/&gt;
  &lt;/button&gt;
&lt;/dml&gt;
</pre>

###END###

*****************************************************************************/

static ERROR GET_Template(struct ScrollBar *Self, STRING *Value)
{
   if (Self->Template) {
      *Value = Self->Template;
      return(ERR_Okay);
   }
   else return(ERR_FieldNotSet);
}

static ERROR SET_Template(struct ScrollBar *Self, STRING Value)
{
   WORD i;

   if (Value) {
      for (i=0; (i < sizeof(Self->Template)-1) AND (Value[i]); i++) Self->Template[i] = Value[i];
      Self->Template[i] = 0;
   }
   else Self->Template[0] = 0;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Width
Short:  Defines the width of a scrollbar.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>A scrollbar can be given a fixed or relative width by setting this field
to the desired value.  To set a relative width, use the FD_PERCENT flag when
setting the field.</p>

###END###

*****************************************************************************/

static ERROR GET_Width(struct ScrollBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetWidth, ERR_AccessObject));
}

static ERROR SET_Width(struct ScrollBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (((Value->Type & FD_DOUBLE) AND (!Value->Double)) OR
       ((Value->Type & FD_LARGE) AND (!Value->Large))) {
      return(ERR_Okay);
   }

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_Width, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   XCoord
Short:  The horizontal position of a scrollbar.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The horizontal position of a scrollbar can be set to an absolute or relative
coordinate by writing a value to the XCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Field: XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_X(struct ScrollBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_XCoord, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetX, ERR_AccessObject));
}

static ERROR SET_X(struct ScrollBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_XCoord, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   XOffset
Short:  The horizontal offset of a scrollbar.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The XOffset has a dual purpose depending on whether or not it is set in
conjunction with an X coordinate or a Width based field.</p>

<p>If set in conjunction with an X coordinate then the scrollbar will be
drawn from that X coordinate up to the width of the container, minus the
value given in the XOffset.  This means that the width of the ScrollBar is
dynamically calculated in relation to the width of the container.</p>

<p>If the XOffset field is set in conjunction with a fixed or relative width
then the scrollbar will be drawn at an X coordinate calculated from the formula
&quot;XCoord = ContainerWidth - ScrollBarWidth - XOffset&quot;.</p>

###SEE ALSO###
Field: XCoord, Width, YOffset

###END###

*****************************************************************************/

static ERROR GET_XOffset(struct ScrollBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_XOffset, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetXOffset, ERR_AccessObject));
}

static ERROR SET_XOffset(struct ScrollBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_XOffset, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   YCoord
Short:  The vertical position of a scrollbar.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The vertical position of a ScrollBar can be set to an absolute or relative
coordinate by writing a value to the YCoord field.  To set a
relative/percentage based value, you must use the FD_PERCENT flag or the
value will be interpreted as fixed.  Negative values are permitted.</p>

###SEE ALSO###
Field: YOffset, XCoord

###END###

*****************************************************************************/

static ERROR GET_Y(struct ScrollBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_YCoord, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetY, ERR_AccessObject));

}

static ERROR SET_Y(struct ScrollBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_YCoord, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

/*****************************************************************************

###FIELD###
Name:   YOffset
Short:  The vertical offset of a scrollbar.
Type:   DOUBLE/PERCENTAGE
Status: Get/Set

###DESCRIPTION###
<p>The YOffset has a dual purpose depending on whether or not it is set in
conjunction with a Y coordinate or a Height based field.</p>

<p>If set in conjunction with a Y coordinate then the scrollbar will be drawn
from that Y coordinate up to the height of the container, minus the value
given in the YOffset.  This means that the height of the scrollbar is
dynamically calculated in relation to the height of the container.</p>

<p>If the YOffset field is set in conjunction with a fixed or relative height
then the scrollbar will be drawn at a Y coordinate calculated from the formula
&quot;YCoord = ContainerHeight - ScrollBarHeight - YOffset&quot;.</p>

###SEE ALSO###
Field: YCoord, Height, XOffset

###END###

*****************************************************************************/

static ERROR GET_YOffset(struct ScrollBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_YOffset, FT_DOUBLE, &value);
      ReleaseObject(drawable);

      if (Value->Type & FD_DOUBLE) Value->Double = value;
      else if (Value->Type & FD_LARGE) Value->Large = value;
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetYOffset, ERR_AccessObject));
}

static ERROR SET_YOffset(struct ScrollBar *Self, struct Variable *Value)
{
   OBJECTPTR drawable;

   if (AccessObject(Self->RegionID, 4000, &drawable) IS ERR_Okay) {
      SetField(drawable, FID_YOffset, FT_VARIABLE, Value);
      ReleaseObject(drawable);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

