/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      CommandLine
Version:   1.0
ID:        ID_COMMANDLINE
Category:  Tool
Date:      January 2004
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2000-2004.  All rights reserved.
Keywords:  command, line, cli, interface, user, text, terminal, shell
Short:     Provides a command-line interface between the user and the system.

###DESCRIPTION###
<p>The CommandLine class is a tool that can be used to create
text-based Command Line Interfaces (CLI's).  Creating a valid CommandLine
object only requires a reference to an object capable of receiving text based
output (usually a Text object).  Whenever a command is executed, the
CommandLine object will process it and then send the results to the output
object you specify.  For example, the command "print athene:documents/readme.txt"
would send the readme.txt file to the Output object as text information.</p>

<pIn order to execute a command, the request needs to be written to the
Command field and then the CommandLine object needs to be Activated for
the command to be executed.  This means that you will need to set up an
interface mechanism for sending information to the CommandLine object
(using the input template is a satisfactory solution).</p>

<p>If an invalid command is used, the CommandLine object will print a human
readable error to the Output object.  This ensures that feedback is always
sent to the user even when there is no success.</p>

<p>A number of methods are provided by this class which are useful for
creating menu-based interfaces to CommandLine objects.  For further
information as to the arguments required for the execution of these
methods, refer to the relevant documented sections.</p>

<p>This document does not describe any of the available commands supported by
this class (except those also supported via methods).  If you need
information on the commands available to the user, open the Athene CLI
application and type 'help' for an index.  It is also recommended that you
view the DML code to the CLI application, which you can find in the
scripts:tools/commandline/ directory.</p>

###END###

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/modules/strings.h>
#include <pandora/modules/filesystem.h>
#include <pandora/tools/commandline.h>
#include <pandora/data/config.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_COMMANDLINE 1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_COMMANDLINE, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2004.  All rights reserved.",
   "January 2004",
   "CommandLine"
};

static struct KernelBase *KernelBase;
static struct StringsBase *StringsBase;
static struct FileSystemBase *FileSystemBase;
static OBJECTPTR modFileSystem = NULL, modStrings = NULL;
static OBJECTPTR CommandLineClass = 0;
static OBJECTPTR glCommands = 0;

static BYTE Help[] = {
"The following commands are internal to the Command Line Interface:\n\
\n\
   about   - General information about the CLI.\n\
   create  - Creates an object from a class of your choosing.\n\
   erase   - Destroys all objects created from the CLI.\n\
   expunge - Clears the system of redundant program code.\n\
   free    - Destroy any object by ID or name.\n\
   script  - Run an object script (e.g. DML).\n\
   status  - Lists the objects that you have created through the CLI.\n"
};

static BYTE About[] = {
"Command Line Interface V1.0\n\
\n\
Copyright Rocklyte Systems (c) 2000-2004.  All rights reserved.\n\
\n\
\n\
INTRODUCTION\n\
\n\
The Command Line Interface, or 'CLI' provides you with a direct\n\
interface to the Athene system, including a number of powerful\n\
commands that are not available from the GUI interface.  You can\n\
use the CLI to run programs, launch scripts, create objects and\n\
manipulate the file system amongst other things.  To obtain\n\
a list of commands that are available to you, enter 'help'.\n\
\n\
\n\
COMMANDS\n\
\n\
The commands that are available from the CLI depend on the files\n\
that have been installed on your system.  There are three types\n\
of programs that you can execute from the CLI: Executables,\n\
Command Classes and Command Scripts.\n\
\n\
Executables can be launched by using the run command, but you need\n\
to know the exact location of an executable to run it.  For\n\
instance, \"run src=\"programs:demos/moire\".\n\
\n\
Command classes are installed to the system:classes/commands/\n\
directory and are included in the help list.  You can write your\n\
own command classes if you have the Pandora Engine SDK.\n\
\n\
Command scripts are installed in the scripts:commands/ directory\n\
and are also included in the help list.  You can write your own\n\
command scripts if you learn an object scripting language like\n\
DML.\n"
};

static BYTE CreateHelp[] = {
"Usage: CREATE [CLASS] [ARGS...]\n\
\n\
Using the create command will create and initialise a new object\n\
from the class of your choosing.  In order to use this command\n\
effectively, you will need to have read about or developed an\n\
understanding of the class that you wish to use first.\n\
\n\
This example will draw a circle to the CLI:\n\
\n\
  create ellipse x=50% y=50% width=10 height=10 colour=\"#000000\"\n\
\n\
A record of static objects that originate from the create command\n\
can be recalled by using the 'status' command.\n"
};

static BYTE EraseHelp[] = {
"Usage: ERASE\n\
\n\
Using the erase command will cause all objects that you have\n\
created (using the 'create' command) to be deleted from the\n\
system.\n"
};

static BYTE ExpungeHelp[] = {
"Usage: EXPUNGE\n\
\n\
Using the expunge command will clear the system of redundant\n\
program code (code that is stored in memory but is not in use\n\
at the time the expunge command is entered).  This can free\n\
up large chunks of system memory if there are Tasks in the\n\
system that have loaded code that is not longer being used.\n\
\n\
Expunging is automatically performed when the system runs\n\
low on memory, but forcing an expunge still has its\n\
usefulness if run periodically.\n"
};

static BYTE FreeHelp[] = {
"Usage: FREE [ID|NAME]\n\
\n\
This powerful command can free any public object from the\n\
system, with no exceptions.  You can use it to not only\n\
kill applications, but individual objects and functions\n\
within any application and the OS.\n"
};

static UBYTE InstanceHelp[] = {
"Usage: INSTANCE\n\
\n\
Use the instance command if you need to know the instance ID\n\
of the environment that Athene is running in.  Generally you\n\
only ever need to know the instance number when you are\n\
debugging programs.\n"
};

static BYTE ScriptHelp[] = {
"Usage: SCRIPT [FILE] [ARGS...]\n\
\n\
You can run object scripts with this command.  If you want to\n\
specify arguments to be passed to the script, do so after\n\
giving the file location.\n\
\n\
Example:\n\
\n\
  script templates:window.dml width=200 height=100\n"
};

static BYTE StatusHelp[] = {
"Usage: STATUS\n\
\n\
This command prints a list of the objects that you have created\n\
in the CLI.  If you have not created any objects, the command\n\
does nothing.\n"
};

/*****************************************************************************
** Class definition.
*/

#define TABLE_SIZE  100
#define SIZE_BUFFER  4096
#define LEN_COMMAND 300

struct CommandLine {
   OBJECT_HEADER
   OBJECTID OutputID;

   PRIVATE_FIELDS
   OBJECTID ObjectTable[TABLE_SIZE];
   BYTE     Buffer[SIZE_BUFFER];
   BYTE     Command[LEN_COMMAND];
};

static ERROR GET_Command(struct CommandLine *, STRING *);

static ERROR SET_Command(struct CommandLine *, STRING);

static struct FieldArray CommandLineFields[] = {
   { "Output",  0, FDF_OBJECTID|FDF_RW, 0, NULL, NULL },
   /*** Virtual fields ***/
   { "Command", 0, FDF_STRING|FDF_RW, 0, GET_Command, SET_Command },
   END_FIELD
};

/*****************************************************************************
** CommandLine Actions.
*/

static ERROR CLI_Activate(struct CommandLine *, APTR);
static ERROR CLI_Init(struct CommandLine *, APTR);

static struct ActionArray CommandLineActions[] = {
   { AC_Activate, CLI_Activate },
   { AC_Init,     CLI_Init },
   { NULL, NULL }
};

/*****************************************************************************
** CommandLine Methods.
*/

static ERROR CLI_About(struct CommandLine *, APTR);
static ERROR CLI_Erase(struct CommandLine *, APTR);
static ERROR CLI_Help(struct CommandLine *, struct mtCommandHelp *);
static ERROR CLI_Status(struct CommandLine *, APTR);

static struct MethodArray CommandLineMethods[] = {
   { MTC_About,  CLI_About,  "About",  NULL },
   { MTC_Erase,  CLI_Erase,  "Erase",  NULL },
   { MTC_Help,   CLI_Help,   "Help",   NULL },
   { MTC_Status, CLI_Status, "Status", NULL },
   { NULL, NULL, NULL, NULL }
};

/****************************************************************************/

struct Command {
   STRING Name;
   STRING Help;
   ERROR (*Routine)(struct CommandLine *, STRING);
};

static ERROR AboutCommand(struct CommandLine *, STRING);
static ERROR CreateCommand(struct CommandLine *, STRING);
static ERROR ExpungeCommand(struct CommandLine *, STRING);
static ERROR EraseCommand(struct CommandLine *, STRING);
static ERROR FreeCommand(struct CommandLine *, STRING);
static ERROR HelpCommand(struct CommandLine *, STRING);
static ERROR InstanceCommand(struct CommandLine *, STRING);
static ERROR ScriptCommand(struct CommandLine *, STRING);
static ERROR StatusCommand(struct CommandLine *, STRING);

static struct Command Commands[] = {
   { "about",    About,        AboutCommand },
   { "create",   CreateHelp,   CreateCommand },
   { "expunge",  ExpungeHelp,  ExpungeCommand },
   { "erase",    EraseHelp,    EraseCommand },
   { "free",     FreeHelp,     FreeCommand },
   { "help",     Help,         HelpCommand },
   { "instance", InstanceHelp, InstanceCommand },
   { "script",   ScriptHelp,   ScriptCommand },
   { "status",   StatusHelp,   StatusCommand },
   { NULL, NULL, NULL }
};

/****************************************************************************/

static void OutputText(struct CommandLine *, STRING);
static void OutputLine(struct CommandLine *, STRING);
static void ExtractValue(STRING str, STRING Value, LONG BufferLen);
static void WriteField(OBJECTPTR Object, const STRING FieldName, STRING Value);
static void SetVField(OBJECTPTR Object, STRING FieldName, STRING Value);
static void WriteField(OBJECTPTR Object, const STRING FieldName, STRING Value);
static void ProcessFields(struct CommandLine *, STRING, OBJECTPTR);
static void CleanupObjectTable(struct CommandLine *);
static LONG ExtractField(STRING FieldName, STRING str);
static LONG NextElement(STRING str);
static OBJECTID GrabObject(STRING);

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   OBJECTPTR config;

   KernelBase = argKernelBase;

   /*** Open required modules ***/

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &modFileSystem, NULL,
      FID_Name|TSTRING,   "filesystem",
      FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
      TAGEND) IS ERR_Okay) {
      if (GetField(modFileSystem, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   /* Load the classes reference file and filter everything
   ** out that isn't a command.  This procedure can be skipped (but
   ** it will not be possible to execute foreign commands).
   */

   if (CreateObject(ID_CONFIG, NULL, &config, NULL,
         FID_Location|TSTRING,   "system:config/kernel/classes.cfg",
         FID_ItemFilter|TSTRING, "category=command",
         TAGEND) IS ERR_Okay) {
      glCommands = config;
   }

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&CommandLineClass, NULL,
      FID_BaseClassID|TLONG, ID_COMMANDLINE,
      FID_Version|TFLOAT,    VER_COMMANDLINE,
      FID_Name|TSTRING,      "CommandLine",
      FID_Category|TLONG,    CCF_TOOL,
      FID_Actions|TPTR,      CommandLineActions,
      FID_Methods|TPTR,      CommandLineMethods,
      FID_Fields|TPTR,       CommandLineFields,
      FID_Size|TLONG,        sizeof(struct CommandLine),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (glCommands)       { Action(AC_Free, glCommands, NULL);       glCommands = NULL; }
   if (modFileSystem)    { Action(AC_Free, modFileSystem, NULL);    modFileSystem = NULL; }
   if (modStrings)       { Action(AC_Free, modStrings, NULL);       modStrings = NULL; }
   if (CommandLineClass) { Action(AC_Free, CommandLineClass, NULL); CommandLineClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  About
Short: Prints summarised information about the CommandLine class.

###DESCRIPTION###
<p>If this method is called the CommandLine object will print a summary of
what the CommandLine does, who wrote it, copyright information and so on.</p>

###RESULT###
ERR_Okay: The method executed successfully.
ERR_FieldNotSet: The Output field has not been set.

###END###

*****************************************************************************/

static ERROR CLI_About(struct CommandLine *Self, APTR Void)
{
   OutputText(Self, About);
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Activate
Short: Executes the instructions set in a CommandLine's Command field.

###DESCRIPTION###
<p>Activating a CommandLine object causes it to analyse the
<a href="#Command">Command</a> field and print the results to the object
specified in the <a href="#Output">Output</a> field.</p>

###END###

*****************************************************************************/

static ERROR CLI_Activate(struct CommandLine *Self, APTR Void)
{
   struct acSetUnlistedField unlisted;
   struct mtReadConfig read;
   struct FileInfo *list, *scan;
   UBYTE FieldName[60], Value[600], command[40];
   OBJECTPTR object, script;
   CLASSID ClassID;
   STRING str;
   LONG i, j, len;
   ERROR error;

   if (!Self->OutputID) {
      DPrintF("!Activate()","[CommandLine:%d] The Output field has not been set.", GetUniqueID(Self));
      return(ERR_FieldNotSet);
   }
   else if ((!Self->Command) OR (!Self->Command[0])) {
      DPrintF("!Activate()","[CommandLine:%d] The Command field has not been set.", GetUniqueID(Self));
      return(ERR_FieldNotSet);
   }

   for (len=0; (Self->Command[len]) AND (Self->Command[len] != ' ') AND (len < sizeof(command)-1); len++) {
      command[len] = Self->Command[len];
   }
   command[len] = 0;

   for (i=0; Commands[i].Name; i++) {
      if (StrCompare(Commands[i].Name, command, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
         while (Self->Command[len] IS ' ') len++;
         if (!Self->Command[len]) error = Commands[i].Routine(Self, NULL);
         else error = Commands[i].Routine(Self, Self->Command + len);

         /*** If an ERR_Args code was returned, print the help for the command ***/

         if (error IS ERR_Args) OutputText(Self, Commands[i].Help);

         return(ERR_Okay);
      }
   }

   if (!Commands[i].Name) {
      /* Search the class registry for command classes that might match the
      ** command name that has been entered.  If one is found, it is used
      ** to execute the command line.
      */

      read.Section = command;
      read.Item    = "Location";
      if (Action(MT_ReadConfig, glCommands, &read) IS ERR_Okay) {
         ActionMsg(AC_Clear, Self->OutputID, NULL);
         if ((ClassID = ResolveClassName(command))) {
            if (NewPrivateObject(ClassID, NULL, &object) IS ERR_Okay) {
               ProcessFields(Self, Self->Command + len, object);

               if (Action(AC_Init, object, NULL) IS ERR_Okay) {
                  Action(AC_Activate, object, NULL);
                  Action(AC_Free, object, NULL);
                  return(ERR_Okay);
               }
               else {
                  Action(AC_Free, object, NULL);
                  OutputText(Self, "Failed to initialise the object for this command (try 'help command').");
                  return(ERR_Init);
               }
            }
            else {
               OutputText(Self, "Failed to create the command's relevant object.");
               return(ERR_NewObject);
            }
         }
         else {
            OutputText(Self, "Could not load the command's class.");
            return(ERR_Search);
         }
      }

      /* Get a directory list of the commands: directory and look
      ** for a file that matches the typed instruction.
      */

      if (ReadDirectory("commands:", RDF_FILE, &list) IS ERR_Okay) {
         for (scan=list; scan; scan=scan->Next) {
            for (j=0; (scan->Name[j]) AND (scan->Name[j] != '.'); j++);
            if ((j IS len) AND (StrCompare(command, scan->Name, j, NULL) IS ERR_Okay)) {
               if (NewPrivateObject(ID_SCRIPT, NULL, &script) IS ERR_Okay) {
                  StrFormat(Self->Buffer, SIZE_BUFFER, "commands:%s", scan->Name);

                  SetFields(script, FID_Location|TSTRING, Self->Buffer,
                                    FID_Output|TLONG,     Self->OutputID,
                                    TAGEND);

                  str = Self->Command + len;
                  while (*str) {
                     while (*str IS ' ') str++;
                     if (!*str) break;
                     str += ExtractField(FieldName, str);
                     ExtractValue(str, Value, sizeof(Value));
                     unlisted.Field = FieldName;
                     unlisted.Value = Value;
                     Action(AC_SetUnlistedField, script, &unlisted);
                     str += NextElement(str);
                  }

                  if (Action(AC_Init, script, NULL) IS ERR_Okay) {
                     ActionMsg(AC_Clear, Self->OutputID, NULL);
                     Action(AC_Activate, script, NULL);
                  }
                  Action(AC_Free, script, NULL);
               }
               FreeMemory(list);
               return(ERR_Okay);
            }
         }
         FreeMemory(list);
      }

      /*** The command that was entered is invalid ***/

      i = StrCopy("Unsupported command: ", Self->Buffer, COPY_ALL);
      StrCopy(Self->Command, Self->Buffer+i, 60);
      OutputText(Self, Self->Buffer);
      return(ERR_Okay);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  Erase
Short: Destroys all objects created from the CommandLine.

###DESCRIPTION###
<p>Calling the Erase method will destroy every object that has been
instantiated through use of the 'create' command.  This is useful if the
user has created a number of objects and wants to clean them away to start
afresh.</p>

###END###

*****************************************************************************/

static ERROR CLI_Erase(struct CommandLine *Self, APTR Void)
{
   OBJECTPTR Object;
   LONG i, count;
   STRING ClassName;

   CleanupObjectTable(Self);

   count = 0;
   for (i=0; Self->ObjectTable[i]; i++) {
      if (Self->ObjectTable[i]) {
         if (AccessObject(Self->ObjectTable[i], 1000, &Object) IS ERR_Okay) {
            if (count IS 0) OutputText(Self, "Erasing objects...\n");

            GetField(Object->Class, FID_Name, FT_POINTER, &ClassName);

            StrFormat(Self->Buffer, SIZE_BUFFER, "%s: %d", ClassName, Self->ObjectTable[i]);

            Action(AC_Free, Object, NULL);
            Self->ObjectTable[i] = NULL;
            ReleaseObject(Object);
            count++;

            OutputLine(Self, Self->Buffer);
         }
      }
   }

   if (!count) OutputText(Self, "There are no objects to erase.\n");

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  Help
Short: Prints the user instructions for using the Command Line Interface.

###DESCRIPTION###
<p>Calling this method will output the instructions on how the user should
use the CommandLine class.</p>

###END###

*****************************************************************************/

static ERROR CLI_Help(struct CommandLine *Self, struct mtCommandHelp *Args)
{
   struct mtReadConfig read;
   struct ConfigEntry *entries;
   struct FileInfo *list, *scan;
   LONG amtentries, i, j, k, len, total, columns, pos;
   OBJECTPTR class;
   STRING usage, *array;

   if (!Self->OutputID) return(ObjectError(Self, MTC_Help, ERR_FieldNotSet));

   if ((!Args) OR (!Args->Command) OR (!*Args->Command)) {
      OutputText(Self, Help);
      total = 0;
      j = 0;
      if (glCommands) {
         if (GetFields(glCommands, FID_AmtEntries|TLONG, &amtentries,
                                   FID_Entries|TPTR,     &entries,
                                   TAGEND) IS ERR_Okay) {
            for (i=0; i < amtentries; i++) {
               if (StrCompare("Category", entries[i].Item, NULL, STR_MATCHLENGTH) IS ERR_Okay) {
                  if (j >= SIZE_BUFFER) break;
                  for (k=0; (entries[i].Section[k]) AND (j < SIZE_BUFFER-1); k++) {
                     if ((entries[i].Section[k] >= 'A') AND (entries[i].Section[k] <= 'Z')) {
                        Self->Buffer[j++] = entries[i].Section[k] - 'A' + 'a';
                     }
                     else Self->Buffer[j++] = entries[i].Section[k];
                  }
                  Self->Buffer[j++] = 0;
                  total++;
               }
            }
         }
      }

      /*** List the command files in the commands: directory ***/

      if (ReadDirectory("commands:", RDF_FILE, &list) IS ERR_Okay) {
         for (scan=list; scan; scan=scan->Next) {
            for (len=0; (scan->Name[len]) AND (scan->Name[len] != '.'); len++);
            if (len > 0) {
               if (j >= SIZE_BUFFER) break;
               for (k=0; (scan->Name[k]) AND (scan->Name[k] != '.') AND (j < SIZE_BUFFER-1); k++) {
                  Self->Buffer[j++] = lcase(scan->Name[k]);
               }
               Self->Buffer[j++] = 0;
               total++;
            }
         }
         FreeMemory(list);
      }

      if (total > 0) {
         if ((array = StrBuildArray(Self->Buffer, j, total, SBF_NODUPLICATES))) {
            OutputLine(Self, "The following commands are installed on your system:\n");

            /*** Get the length of the longest command name ***/

            len = 0;
            for (i=0; array[i]; i++) {
               for (k=0; array[i][k]; k++);
               if (k > len) len = k;
            }
            len += 3;

            /*** Write the command names in columns ***/

            columns = 70 / len;

            for (i=0; i < total;) {
               pos = 0;
               for (j=0; (j < columns) AND (i < total); j++) {
                  k = pos;
                  Self->Buffer[pos++] = ' ';
                  Self->Buffer[pos++] = ' ';
                  Self->Buffer[pos++] = ' ';
                  pos += StrCopy(array[i], Self->Buffer + pos, COPY_ALL);

                  while ((pos - k) < len) Self->Buffer[pos++] = ' ';
                  i++;
               }
               Self->Buffer[pos++] = 0;
               OutputLine(Self, Self->Buffer);
            }

            FreeMemory(array);
         }
      }

      OutputLine(Self, "\nTo get help on a particular command, type \"help <commandname>\".");
      return(ERR_Okay);
   }
   else {
      for (i=0; Commands[i].Name; i++) {
         if (StrCompare(Commands[i].Name, Args->Command, NULL, NULL) IS ERR_Okay) {
            OutputText(Self, Commands[i].Help);
            return(ERR_Okay);
         }
      }

      read.Section = Args->Command;
      read.Item    = "Location";
      if (Action(MT_ReadConfig, glCommands, &read) IS ERR_Okay) {
         if ((class = FindClass(ResolveClassName(Args->Command), NULL))) {
            if ((GetField(class, FID_Usage, FT_POINTER, &usage) IS ERR_Okay) AND (usage)) {
               OutputText(Self, usage);
            }
            else OutputText(Self, "Help is not available for this command.");

            return(ERR_Okay);
         }
         OutputText(Self, "Could not retrieve help for the command.");
         return(ERR_Failed);
      }
      else {
         SET_Command(Self, Args->Command);
         if (Action(AC_Activate, Self, NULL) IS ERR_Okay) return(ERR_Okay);
         else {
            OutputText(Self, "The command is not known.");
            return(ERR_Search);
         }
      }
   }
}

/*****************************************************************************
** CommandLine: Init
*/

static ERROR CLI_Init(struct CommandLine *Self, APTR Void)
{
   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  Status
Short: Prints a list of all the objects created from the CommandLine.

###DESCRIPTION###
<p>Calling the Status method prints a list of all the objects that have been
created by the CommandLine object.  This allows the user to know exactly
what is lingering in memory as a result of using the 'create' command.
All objects listed by the Status method are freed from the system when the
CommandLine object terminates, or the 'erase' command can be used to destroy
them early.</p>

###END###

*****************************************************************************/

static ERROR CLI_Status(struct CommandLine *Self, APTR Void)
{
   OBJECTPTR Object;
   LONG i, count;
   STRING ClassName;

   count = 0;
   for (i=0; Self->ObjectTable[i]; i++) {
      if (Self->ObjectTable[i]) {
         if (AccessObject(Self->ObjectTable[i], 1000, &Object) IS ERR_Okay) {
            if (count IS 0) OutputText(Self, "Created Objects:\n\n");
            GetField(Object->Class, FID_Name, FT_POINTER, &ClassName);
            StrFormat(Self->Buffer, SIZE_BUFFER, "%s: %d", ClassName, Self->ObjectTable[i]);
            OutputLine(Self, Self->Buffer);
            ReleaseObject(Object);
            count++;
         }
      }
   }

   if (!count) OutputText(Self, "There are no objects owned by the CLI.\n");

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Command
Short:  Set this field prior to activation for a command to be executed.
Status: Read/Write
Type:   STRING

###DESCRIPTION###
<p>Before a CommandLine object is Activated, this field must be set so that
the object will know what command to perform.  Setting this field will not
result in a validation of the command - it will simply buffer the request and
translate it once the Activate action is called on the CommandLine object.</p>

<p>To get a list of valid commands, type 'help' into the Athene CLI
application.</p>

###END###

*****************************************************************************/

static ERROR GET_Command(struct CommandLine *Self, STRING *Value)
{
   *Value = Self->Command;
   return(ERR_Okay);
}

static ERROR SET_Command(struct CommandLine *Self, STRING Value)
{
   LONG i;

   if (!Value) {
      Self->Command[0] = 0;
      return(ERR_Okay);
   }

   for (i=0; (Value[0]) AND (i < sizeof(Self->Command)-1); i++) Self->Command[i] = Value[i];
   Self->Command[i] = 0;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Output
Short:  Set this field to give the CommandLine an object to send output to.
Status: Read/Write
Type:   OBJECTID

###DESCRIPTION###
<p>The Output field must be set in order for a CommandLine object to
process the commands sent from the user.  If it is not set then commands
will not be executed when the CommandLine object is Activated.  The Output
object must accept text-based information via data channels in order for the
data to be interpreted correctly.  If the Output object does not support
text based information, the CommandLine will still process the command but
the user will not receive any text based output.</p>

<p>It is recommended that this field points to an object that is a member
of the Text class.</p>

###SEE ALSO###
Class: Text

###END###

*****************************************************************************/

/*****************************************************************************
** Function: CleanupObjectTable()
*/

static void CleanupObjectTable(struct CommandLine *Self)
{
   LONG i;
   for (i=0; i < TABLE_SIZE; i++) {
      if (Self->ObjectTable[i]) {
         if (CheckObjectExists(Self->ObjectTable[i], NULL) != ERR_True) {
            Self->ObjectTable[i] = NULL;
         }
      }
   }
}

/****************************************************************************/

static void OutputText(struct CommandLine *Self, STRING Buffer)
{
   struct acDataChannel channel;

   if (Self->OutputID) {
      /*** Send the data to the output object ***/

      ActionMsg(AC_Clear, Self->OutputID, NULL);

      channel.ObjectID     = GetUniqueID(Self);
      channel.DataType     = DATA_TEXT;
      channel.Version      = 1;
      channel.Buffer       = Buffer;
      channel.Size         = StrLength(Buffer);
      channel.TotalEntries = 1;
      ActionMsg(AC_DataChannel, Self->OutputID, &channel);
   }
}

/****************************************************************************/

static void OutputLine(struct CommandLine *Self, STRING Buffer)
{
   struct acDataChannel channel;

   if (Self->OutputID) {
      /*** Send the data to the output object ***/
      channel.ObjectID     = GetUniqueID(Self);
      channel.DataType     = DATA_TEXT;
      channel.Version      = 1;
      channel.Buffer       = Buffer;
      channel.Size         = StrLength(Buffer);
      channel.TotalEntries = 1;
      ActionMsg(AC_DataChannel, Self->OutputID, &channel);
   }
}

/****************************************************************************/

static void ProcessFields(struct CommandLine *Self, STRING str, OBJECTPTR Object)
{
   BYTE FieldName[60], Value[600];

   SetField(Object, FID_Output, FT_LONG, &Self->OutputID);

   while (*str) {
      while (*str IS ' ') str++;
      if (!*str) break;
      str += ExtractField(FieldName, str);
      ExtractValue(str, Value, sizeof(Value));
      WriteField(Object, FieldName, Value);
      str += NextElement(str);
   }
}

/*****************************************************************************
** Function: ExtractValue()
**
** This function can be sent strings such as:
**
**   myvalue = blah
**   myvalue=    blah
**   myvalue="blah blah blah"
**
** All of which would result in "blah" or "blah blah blah" being written to
** the Value buffer.  For convenience, quotes are always removed from the
** resulting value.
**
** If a value such as the following is sent with no = sign after the text:
**
**   myvalue
**
**  Then the value will be set to a default of TRUE (1).
*/

static void ExtractValue(STRING str, STRING Value, LONG BufferLen)
{
   LONG i;

   Value[0] = '1';
   Value[1] = 0;

   if (*str != '=') {
      while ((*str != ' ') AND (*str != '=') AND (*str != 0)) str++;
      while (*str IS ' ') str++;
   }

   if (*str IS '=') {
      str++;
      while (*str IS ' ') str++;
      if (*str IS '"') {
         str++;
         for (i=0; (*str) AND (*str != '"'); i++) {
            Value[i] = *str++;
            if (i >= (BufferLen - 1)) { DPrintF("!CommandLine:","Buffer overflow."); break; }
         }
         if (*str) str++; /* Skip the final quote */
      }
      else if (*str IS '\'') {
         str++;
         for (i=0; (*str) AND (*str != '\''); i++) {
            Value[i] = *str++;
            if (i >= (BufferLen - 1)) { DPrintF("!CommandLine:","Buffer overflow."); break; }
         }
         if (*str) str++; /* Skip the final quote */
      }
      else {
         for (i=0; (*str) AND (*str != ' '); i++) {
            Value[i] = *str++;
            if (i >= (BufferLen - 1)) { DPrintF("!CommandLine:","Buffer overflow."); break; }
         }
      }
      Value[i] = 0;
   }
}

/*****************************************************************************
** Function: WriteField()
*/

static void WriteField(OBJECTPTR Object, const STRING FieldName, STRING Value)
{
   OBJECTPTR Context;
   BYTE Translation[256];
   LONG len;

   if ((!Object) OR (!FieldName) OR (!Value)) return;

   Context = SetContext(Object);

      /* If the value is encapsulated with dollar symbols '$' then we
      ** do not translate it (we just remove the dollar symbols).
      */

      for (len=0; Value[len]; len++);
      if ((Value[0] IS '$') AND (Value[len-1] IS '$')) {
         Value[len-1] = 0;
         SetVField(Object, FieldName, Value+1);
      }
      else {
         StrCopy(Value, Translation, COPY_ALL);
         StrTranslate(Translation, sizeof(Translation), NULL);
         SetVField(Object, FieldName, Translation);
     }

   SetContext(Context);
}

static void SetVField(OBJECTPTR Object, STRING FieldName, STRING Value)
{
   DPrintF("CLI:","Writing value \"%s\" to field \"%s\".", Value, FieldName);

   if (FieldName[0] IS '&') {
      struct acSetUnlistedField UnlistedArgs;
      UnlistedArgs.Field = FieldName+1;
      UnlistedArgs.Value = Value;
      Action(AC_SetUnlistedField, Object, &UnlistedArgs);
   }
   else SetFieldVariable(Object, FieldName, Value);
}

/*****************************************************************************
** Function: NextElement()
*/

static LONG NextElement(STRING str)
{
   LONG i;

   i = 0;
   while (str[i] IS ' ') i++;
   if (str[i] IS '=') {
      i++;
      while (str[i] IS ' ') i++;
      if (str[i] IS '"') {
         i++;
         while ((str[i]) AND (str[i] != '"')) i++;
         if (str[i]) i++;
      }
      else if (str[i] IS '\'') {
         i++;
         while ((str[i]) AND (str[i] != '\'')) i++;
         if (str[i]) i++;
      }
      else while ((str[i]) AND (str[i] != ' ')) i++;
   }
   return(i);
}

static LONG ExtractField(STRING FieldName, STRING str)
{
   LONG i;

   i = 0;
   while (str[i] IS ' ') i++;
   while ((str[i]) AND (str[i] != '=') AND (str[i] != ' ')) {
      *FieldName++ = str[i++];
   }
   *FieldName = 0;
   return(i);
}

/****************************************************************************/

static OBJECTID GrabObject(STRING Name)
{
   OBJECTID ObjectID;
   LONG count;

   if ((*Name IS '-') OR ((*Name >= '0') AND (*Name <= '9'))) {
      ObjectID = (OBJECTID)StrToInt(Name);
      if (ObjectID > 0) ObjectID = -ObjectID;
   }

   if (FastFindObject(Name, NULL, &ObjectID, 1, &count) IS ERR_Okay) {
      return(ObjectID);
   }
   else return(NULL);
}

#include "commands.c"
